# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path
from typing import Any, Optional, Union

import click
from lightning_cloud.openapi.rest import ApiException

from lightning.app.cli.cmd_clusters import _check_cluster_id_is_valid, AWSClusterManager
from lightning.app.cli.cmd_ssh_keys import _SSHKeyManager


@click.group("create")
def create() -> None:
    """Create Lightning AI self-managed resources (clusters, etc…)"""
    pass


@create.command("cluster")
@click.argument("cluster_id", callback=_check_cluster_id_is_valid)
@click.option("--provider", "provider", type=str, default="aws", help="cloud provider to be used for your cluster")
@click.option("--external-id", "external_id", type=str, required=True)
@click.option(
    "--role-arn", "role_arn", type=str, required=True, help="AWS role ARN attached to the associated resources."
)
@click.option(
    "--region",
    "region",
    type=str,
    required=False,
    default="us-east-1",
    help="AWS region that is used to host the associated resources.",
    hidden=True,
)
@click.option(
    "--enable-performance",
    "enable_performance",
    type=bool,
    required=False,
    default=False,
    hidden=True,
    is_flag=True,
    help=""""Use this flag to ensure that the cluster is created with a profile that is optimized for performance.
        This makes runs more expensive but start-up times decrease.""",
)
@click.option(
    "--edit-before-creation",
    default=False,
    is_flag=True,
    hidden=True,
    help="Edit the cluster specs before submitting them to the API server.",
)
@click.option(
    "--sync",
    "do_sync",
    type=bool,
    required=False,
    default=False,
    is_flag=True,
    help="This flag makes the CLI wait until cluster creation completes.",
)
def create_cluster(
    cluster_id: str,
    region: str,
    role_arn: str,
    external_id: str,
    provider: str,
    edit_before_creation: bool,
    enable_performance: bool,
    do_sync: bool,
    **kwargs: Any,
) -> None:
    """Create a Lightning AI BYOC compute cluster with your cloud provider credentials."""
    if provider.lower() != "aws":
        click.echo("Only AWS is supported for now. But support for more providers is coming soon.")
        return
    cluster_manager = AWSClusterManager()
    cluster_manager.create(
        cluster_id=cluster_id,
        region=region,
        role_arn=role_arn,
        external_id=external_id,
        edit_before_creation=edit_before_creation,
        cost_savings=not enable_performance,
        do_async=not do_sync,
    )


@create.command("ssh-key")
@click.option("--name", "key_name", default=None, help="name of ssh key")
@click.option("--comment", "comment", default="", help="comment detailing your SSH key")
@click.option(
    "--public-key",
    "public_key",
    help="public key or path to public key file",
    required=True,
)
def add_ssh_key(
    public_key: Union[str, "os.PathLike[str]"], key_name: Optional[str] = None, comment: Optional[str] = None
) -> None:
    """Add a new Lightning AI ssh-key to your account."""
    ssh_key_manager = _SSHKeyManager()

    new_public_key = Path(str(public_key)).read_text() if os.path.isfile(str(public_key)) else public_key
    try:
        ssh_key_manager.add_key(name=key_name, comment=comment, public_key=str(new_public_key))
    except ApiException as e:
        # if we got an exception it might be the user passed the private key file
        if os.path.isfile(str(public_key)) and os.path.isfile(f"{public_key}.pub"):
            ssh_key_manager.add_key(name=key_name, comment=comment, public_key=Path(f"{public_key}.pub").read_text())
        else:
            raise e



from placeholdername import ComponentA, ComponentB

import lightning as L


class LitApp(L.LightningFlow):
    def __init__(self) -> None:
        super().__init__()
        self.component_a = ComponentA()
        self.component_b = ComponentB()

    def run(self):
        self.component_a.run()
        self.component_b.run()


app = L.LightningApp(LitApp())



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import re
import shutil
import subprocess
import sys
import tarfile
import urllib.request
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, List, Optional

import click
from jinja2 import Environment, FileSystemLoader
from rich import print
from rich.panel import Panel
from rich.status import Status
from rich.text import Text
from rich.tree import Tree

import lightning.app

_REPORT_HELP_TEXTS = {
    "core": "Important files for the app such as various components",
    "source": "A copy of all your source code, including the PL script ⚡",
    "tests": "This app comes with tests!",
    "ui": "Source and build files for the user interface",
    "app.py": "This is the main app file!",
    "requirements.txt": "Lists the dependencies required to be installed before running the app",
}

_REPORT_IGNORE_PATTERNS = [
    r"__pycache__",
    r"__init__\.py",
    r".*egg-info",
    r"\..*",
]


def pl_app(source_dir: str, script_path: str, name: str, overwrite: bool) -> None:
    source_dir = Path(source_dir).resolve()
    script_path = Path(script_path).resolve()

    if not source_dir.is_dir():
        click.echo(f"The given source directory does not exist: {source_dir}", err=True)
        raise SystemExit(1)

    if not script_path.exists():
        click.echo(f"The given script path does not exist: {script_path}", err=True)
        raise SystemExit(1)

    if not script_path.is_file():
        click.echo(f"The given script path must be a file, you passed: {script_path}", err=True)
        raise SystemExit(1)

    if source_dir not in script_path.parents:
        click.echo(
            "The given script path must be a subpath of the source directory. Example:"
            " lightning init pl-app ./code ./code/scripts/train.py",
            err=True,
        )
        raise SystemExit(1)

    rel_script_path = script_path.relative_to(source_dir)
    cwd = Path.cwd()
    destination = cwd / name

    if destination.exists():
        if not overwrite:
            click.echo(
                f"There is already an app with the name {name} in the current working directory. Choose a different"
                f" name with `--name` or force to overwrite the existing folder by passing `--overwrite`.",
                err=True,
            )
            raise SystemExit(1)
        else:
            shutil.rmtree(destination)

    template_dir = Path(lightning.app.cli.__file__).parent / "pl-app-template"

    with Status("[bold green]Copying app files"):
        shutil.copytree(template_dir, destination, ignore=shutil.ignore_patterns("node_modules", "build"))
        if (template_dir / "ui" / "build").exists():
            shutil.copytree(template_dir / "ui" / "build", destination / "ui" / "build")
        else:
            download_frontend(destination / "ui" / "build")

    with Status("[bold green]Copying source files"):
        shutil.copytree(source_dir, destination / "source", ignore=shutil.ignore_patterns(name))
        project_file_from_template(template_dir, destination, "app.py", script_path=str(rel_script_path))
        project_file_from_template(template_dir, destination, "setup.py", app_name=name)

    with Status("[bold green]Installing"):
        subprocess.call(["pip", "install", "--quiet", "-e", str(destination)])
        # TODO: download the ui files

    print_pretty_report(
        destination,
        ignore_patterns=_REPORT_IGNORE_PATTERNS,
        help_texts=_REPORT_HELP_TEXTS,
    )


def download_frontend(destination: Path) -> None:
    # TODO: Update the URL to the release in GitHub once the PL app repo is public
    url = "https://storage.googleapis.com/grid-packages/pytorch-lightning-app/v0.0.0/build.tar.gz"
    build_dir_name = "build"
    with TemporaryDirectory() as download_dir:
        response = urllib.request.urlopen(url)
        file = tarfile.open(fileobj=response, mode="r|gz")
        file.extractall(path=download_dir)
        shutil.move(str(Path(download_dir, build_dir_name)), destination)


def project_file_from_template(template_dir: Path, destination_dir: Path, template_name: str, **kwargs: Any) -> None:
    env = Environment(loader=FileSystemLoader(template_dir))
    template = env.get_template(template_name)
    rendered_template = template.render(**kwargs)
    with open(destination_dir / template_name, "w") as file:
        file.write(rendered_template)


def print_pretty_report(
    directory: pathlib.Path,
    ignore_patterns: Optional[List[str]] = None,
    help_texts: Optional[Dict[str, str]] = None,
) -> None:
    """Prints a report for the generated app."""
    tree = Tree(
        f":open_file_folder: [link file://{directory}]{directory}",
        guide_style="bold bright_blue",
    )

    help_texts = {} if help_texts is None else help_texts

    paths = sorted(
        directory.glob("*"),
        key=lambda p: (p.is_file(), p.name.lower()),
    )
    max_witdth = max(len(p.name) for p in paths)

    patterns_to_ignore = [] if ignore_patterns is None else ignore_patterns
    for path in paths:
        if any(re.match(pattern, path.name) for pattern in patterns_to_ignore):
            # Only display relevant files
            continue

        help_text = help_texts.get(path.name, "")
        padding = " " * (max_witdth - len(path.name))

        text_pathname = Text(path.name, "green")
        text_pathname.highlight_regex(r"\..*$", "bold red")
        text_pathname.stylize(f"link file://{path}")
        text_pathname.append(f" {padding} {help_text}", "blue")

        icon = "📂 " if path.is_dir() else "📄 "
        icon = icon if _can_encode_icon(icon) else ""

        tree.add(Text(icon) + text_pathname)

    print("\n")
    print("Done. The app is ready here:\n")
    print(tree)
    print("\nRun it:\n")
    print(Panel(f"[red]lightning run app {directory.relative_to(Path.cwd()) / 'app.py'}"))


def _can_encode_icon(icon: str) -> bool:
    """Helper function to check whether an icon can be encoded."""
    try:
        icon.encode(sys.stdout.encoding)
        return True
    except UnicodeEncodeError:
        return False



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from dataclasses import dataclass


@dataclass
class _Config:
    id = os.getenv("LIGHTNING_USER_ID")
    key = os.getenv("LIGHTNING_API_KEY")
    url = os.getenv("LIGHTNING_CLOUD_URL", "https://lightning.ai")
    api_key = os.getenv("LIGHTNING_API_KEY")
    username = os.getenv("LIGHTNING_USERNAME")
    video_location = os.getenv("VIDEO_LOCATION", "./artifacts/videos")
    har_location = os.getenv("HAR_LOCATION", "./artifacts/hars")
    slowmo = os.getenv("SLOW_MO", "0")



import os.path
from unittest.mock import Mock

import pytest
from core.callbacks import PLAppArtifactsTracker, PLAppProgressTracker, PLAppSummary
from core.components.script_runner import ScriptRunner

from lightning.app.storage import Path
from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.loggers import TensorBoardLogger


@pytest.mark.parametrize("rank", (0, 1))
def test_progress_tracker_enabled(rank):
    trainer = Mock()
    trainer.global_rank = rank
    trainer.is_global_zero = rank == 0
    work = Mock()
    tracker = PLAppProgressTracker(work)
    assert not tracker.is_enabled
    tracker.setup(trainer, Mock(), Mock())
    assert tracker.is_enabled == trainer.is_global_zero


def test_summary_callback_tracks_hyperparameters():
    class ModelWithParameters(LightningModule):
        def __init__(self, float_arg=0.1, int_arg=5, bool_arg=True, string_arg="string"):
            super().__init__()
            self.save_hyperparameters()

    model = ModelWithParameters()
    work = Mock()
    summary = PLAppSummary(work)
    trainer = Trainer(max_epochs=22, callbacks=[summary])  # this triggers the `Callback.on_init_end` hook
    summary.setup(trainer, model)
    assert work.model_hparams == {
        "float_arg": "0.1",
        "int_arg": "5",
        "bool_arg": "True",
        "string_arg": "string",
    }

    assert work.trainer_hparams["max_epochs"] == "22"
    assert work.trainer_hparams["logger"] == "True"
    assert "ModelCheckpoint" in work.trainer_hparams["callbacks"]
    assert "PLAppSummary" in work.trainer_hparams["callbacks"]


def test_artifacts_tracker(tmpdir):
    work = ScriptRunner(root_path=os.path.dirname(__file__), script_path=__file__)
    tracker = PLAppArtifactsTracker(work=work)
    trainer = Mock()

    trainer.loggers = []
    trainer.default_root_dir = "default_root_dir"
    tracker.setup(trainer=trainer, pl_module=Mock())
    assert work.log_dir == Path("default_root_dir")
    assert not work.logger_metadatas

    trainer.loggers = [TensorBoardLogger(save_dir=tmpdir)]
    trainer.logger = trainer.loggers[0]
    tracker.setup(trainer=trainer, pl_module=Mock())
    assert work.log_dir == Path(tmpdir / "lightning_logs" / "version_0")
    assert len(work.logger_metadatas) == 1
    assert work.logger_metadatas[0] == {"class_name": "TensorBoardLogger"}

    # call setup a second time and the metadata length should not change
    tracker.setup(trainer=trainer, pl_module=Mock())
    assert len(work.logger_metadatas) == 1



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from datetime import datetime
from json import JSONDecodeError

from websocket import WebSocketApp

from lightning.app.utilities.app_helpers import Logger

logger = Logger(__name__)


# This is a superclass to inherit log entry classes from it:
# it implements magic methods to sort logs by timestamps.
@dataclass
class _OrderedLogEntry:
    message: str
    timestamp: datetime

    def __ge__(self, other: "_OrderedLogEntry") -> bool:
        return self.timestamp >= other.timestamp

    def __gt__(self, other: "_OrderedLogEntry") -> bool:
        return self.timestamp > other.timestamp


# A general error callback for log reading, prints most common types of possible errors.
def _error_callback(ws_app: WebSocketApp, error: Exception):
    errors = {
        KeyError: "Malformed log message, missing key",
        JSONDecodeError: "Malformed log message",
        TypeError: "Malformed log format",
        ValueError: "Malformed date format",
    }
    logger.error(f"⚡ Error while reading logs ({errors.get(type(error), 'Unknown')}), {error}")
    ws_app.close()



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import inspect
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Type, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from lightning.app import LightningFlow, LightningWork


class LightningVisitor(ast.NodeVisitor):
    """
    Base class for visitor that finds class definitions based on
    class inheritance.
    Derived classes are expected to define class_name and implement
    the analyze_class_def method.
    Attributes
    ----------
    class_name: str
        Name of class to identify, to be defined in subclasses.
    """

    class_name: Optional[str] = None

    def __init__(self):
        self.found: List[Dict[str, Any]] = []

    def analyze_class_def(self, node: ast.ClassDef) -> Dict[str, Any]:
        return {}

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        bases = []
        for base in node.bases:
            if type(base) == ast.Attribute:
                bases.append(base.attr)
            elif type(base) == ast.Name:
                bases.append(base.id)
        if self.class_name in bases:
            entry = {"name": node.name, "type": self.class_name}
            entry.update(self.analyze_class_def(node))
            self.found.append(entry)


class LightningModuleVisitor(LightningVisitor):
    """
    Finds Lightning modules based on class inheritance.
    Attributes
    ----------
    class_name: Optional[str]
        Name of class to identify.
    methods: Set[str]
        Names of methods that are part of the LightningModule API.
    hooks: Set[str]
        Names of hooks that are part of the LightningModule API.
    """

    class_name: Optional[str] = "LightningModule"

    methods: Set[str] = {
        "configure_optimizers",
        "forward",
        "freeze",
        "log",
        "log_dict",
        "print",
        "save_hyperparameters",
        "test_step",
        "test_step_end",
        "test_epoch_end",
        "to_onnx",
        "to_torchscript",
        "training_step",
        "training_step_end",
        "training_epoch_end",
        "unfreeze",
        "validation_step",
        "validation_step_end",
        "validation_epoch_end",
    }

    hooks: Set[str] = {
        "backward",
        "get_progress_bar_dict",
        "manual_backward",
        "manual_optimizer_step",
        "on_after_backward",
        "on_before_zero_grad",
        "on_fit_start",
        "on_fit_end",
        "on_load_checkpoint",
        "on_save_checkpoint",
        "on_pretrain_routine_start",
        "on_pretrain_routine_end",
        "on_test_batch_start",
        "on_test_batch_end",
        "on_test_epoch_start",
        "on_test_epoch_end",
        "on_train_batch_start",
        "on_train_batch_end",
        "on_train_epoch_start",
        "on_train_epoch_end",
        "on_validation_batch_start",
        "on_validation_batch_end",
        "on_validation_epoch_start",
        "on_validation_epoch_end",
        "optimizer_step",
        "optimizer_zero_grad",
        "prepare_data",
        "setup",
        "tbptt_split_batch",
        "teardown",
        "train_dataloader",
        "val_dataloader",
        "test_dataloader",
        "transfer_batch_to_device",
    }


class LightningDataModuleVisitor(LightningVisitor):
    """
    Finds Lightning data modules based on class inheritance.
    Attributes
    ----------
    class_name: Optional[str]
        Name of class to identify.
    methods: Set[str]
        Names of methods that are part of the LightningDataModule API.
    """

    class_name = "LightningDataModule"

    methods: Set[str] = {
        "prepare_data",
        "setup",
        "train_dataloader",
        "val_dataloader",
        "test_dataloader",
        "transfer_batch_to_device",
    }


class LightningLoggerVisitor(LightningVisitor):
    """
    Finds Lightning loggers based on class inheritance.
    Attributes
    ----------
    class_name: Optional[str]
        Name of class to identify.
    methods: Set[str]
        Names of methods that are part of the Logger API.
    """

    class_name = "Logger"

    methods: Set[str] = {"log_hyperparams", "log_metrics"}


class LightningCallbackVisitor(LightningVisitor):
    """
    Finds Lightning callbacks based on class inheritance.
    Attributes
    ----------
    class_name: Optional[str]
        Name of class to identify.
    methods: Set[str]
        Names of methods that are part of the Logger API.
    """

    class_name = "Callback"

    methods: Set[str] = {
        "setup",
        "teardown",
        "on_init_start",
        "on_init_end",
        "on_fit_start",
        "on_fit_end",
        "on_sanity_check_start",
        "on_sanity_check_end",
        "on_train_batch_start",
        "on_train_batch_end",
        "on_train_epoch_start",
        "on_train_epoch_end",
        "on_validation_epoch_start",
        "on_validation_epoch_end",
        "on_test_epoch_start",
        "on_test_epoch_end",
        "on_epoch_start",
        "on_epoch_end",
        "on_batch_start",
        "on_validation_batch_start",
        "on_validation_batch_end",
        "on_test_batch_start",
        "on_test_batch_end",
        "on_batch_end",
        "on_train_start",
        "on_train_end",
        "on_pretrain_routine_start",
        "on_pretrain_routine_end",
        "on_validation_start",
        "on_validation_end",
        "on_test_start",
        "on_test_end",
        "on_keyboard_interrupt",
        "on_save_checkpoint",
        "on_load_checkpoint",
    }


class LightningStrategyVisitor(LightningVisitor):
    """
    Finds Lightning callbacks based on class inheritance.
    Attributes
    ----------
    class_name: Optional[str]
        Name of class to identify.
    methods: Set[str]
        Names of methods that are part of the Logger API.
    """

    class_name = "Strategy"

    methods: Set[str] = {
        "setup",
        "train",
        "training_step",
        "validation_step",
        "test_step",
        "backward",
        "barrier",
        "broadcast",
        "sync_tensor",
    }


class LightningTrainerVisitor(LightningVisitor):
    class_name = "Trainer"


class LightningCLIVisitor(LightningVisitor):
    class_name = "LightningCLI"


class LightningPrecisionPluginVisitor(LightningVisitor):
    class_name = "PrecisionPlugin"


class LightningAcceleratorVisitor(LightningVisitor):
    class_name = "Accelerator"


class LightningLoopVisitor(LightningVisitor):
    class_name = "Loop"


class TorchMetricVisitor(LightningVisitor):
    class_name = "Metric"


class LightningLiteVisitor(LightningVisitor):  # deprecated
    class_name = "LightningLite"


class FabricVisitor(LightningVisitor):
    class_name = "Fabric"


class LightningProfilerVisitor(LightningVisitor):
    class_name = "Profiler"


class Scanner:
    """
    Finds relevant Lightning objects in files in the file system.
    Attributes
    ----------
    visitor_classes: List[Type]
        List of visitor classes to use when traversing files.
    Parameters
    ----------
    path: str
        Path to file, or directory where to look for files to scan.
    glob_pattern: str
        Glob pattern to use when looking for files in the path,
        applied when path is a directory. Default is "**/*.py".
    """

    # TODO: Finalize introspecting the methods from all the discovered methods.
    visitor_classes: List[Type] = [
        LightningCLIVisitor,
        LightningTrainerVisitor,
        LightningModuleVisitor,
        LightningDataModuleVisitor,
        LightningCallbackVisitor,
        LightningStrategyVisitor,
        LightningPrecisionPluginVisitor,
        LightningAcceleratorVisitor,
        LightningLoggerVisitor,
        LightningLoopVisitor,
        TorchMetricVisitor,
        LightningLiteVisitor,  # deprecated
        FabricVisitor,
        LightningProfilerVisitor,
    ]

    def __init__(self, path: str, glob_pattern: str = "**/*.py"):
        path_ = Path(path)
        if path_.is_dir():
            self.paths = path_.glob(glob_pattern)
        else:
            self.paths = [path_]

        self.modules_found: List[Dict[str, Any]] = []

    def has_class(self, cls) -> bool:
        # This method isn't strong enough as it is using only `ImportFrom`.
        # TODO: Use proper classDef scanning.
        classes = []

        for path in self.paths:
            try:
                module = ast.parse(path.open().read())
            except SyntaxError:
                print(f"Error while parsing {path}: SKIPPING")
                continue

            for node in ast.walk(module):

                if isinstance(node, ast.ImportFrom):
                    for import_from_cls in node.names:
                        classes.append(import_from_cls.name)

                if isinstance(node, ast.Call):
                    cls_name = getattr(node.func, "attr", None)
                    if cls_name:
                        classes.append(cls_name)

        return cls.__name__ in classes

    def scan(self) -> List[Dict[str, str]]:
        """
        Finds Lightning modules in files, returning importable
        objects.
        Returns
        -------
        List[Dict[str, Any]]
            List of dicts containing all metadata required
            to import modules found.
        """
        modules_found: Dict[str, List[Dict[str, Any]]] = {}

        for path in self.paths:
            try:
                module = ast.parse(path.open().read())
            except SyntaxError:
                print(f"Error while parsing {path}: SKIPPING")
                continue
            for visitor_class in self.visitor_classes:
                visitor = visitor_class()
                visitor.visit(module)
                if not visitor.found:
                    continue
                _path = str(path)
                ns_info = {
                    "file": _path,
                    "namespace": _path.replace("/", ".").replace(".py", ""),
                }
                modules_found[visitor_class.class_name] = [{**entry, **ns_info} for entry in visitor.found]

        return modules_found


def _is_method_context(component: Union["LightningFlow", "LightningWork"], selected_caller_name: str) -> bool:
    """Checks whether the call to a component originates from within the context of the component's ``__init__``
    method."""
    frame = inspect.currentframe().f_back

    while frame is not None:
        caller_name = frame.f_code.co_name
        caller_self = frame.f_locals.get("self")
        if caller_name == selected_caller_name and caller_self is component:
            # the call originates from a frame under component.__init__
            return True
        frame = frame.f_back

    return False


def _is_init_context(component: Union["LightningFlow", "LightningWork"]) -> bool:
    """Checks whether the call to a component originates from within the context of the component's ``__init__``
    method."""
    return _is_method_context(component, "__init__")


def _is_run_context(component: Union["LightningFlow", "LightningWork"]) -> bool:
    """Checks whether the call to a component originates from within the context of the component's ``run``
    method."""
    return _is_method_context(component, "run") or _is_method_context(component, "load_state_dict")



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import inspect
import os
import pydoc
import subprocess
import sys
from typing import Any, Callable, Optional

import fastapi  # noqa E511
import uvicorn
from fastapi import FastAPI
from fastapi.responses import JSONResponse

from lightning.app.components.serve.types import _DESERIALIZER, _SERIALIZER
from lightning.app.core.work import LightningWork
from lightning.app.utilities.app_helpers import Logger

logger = Logger(__name__)


fastapi_service = FastAPI()


class _InferenceCallable:
    def __init__(
        self,
        deserialize: Callable,
        predict: Callable,
        serialize: Callable,
    ):
        self.deserialize = deserialize
        self.predict = predict
        self.serialize = serialize

    async def run(self, data) -> Any:
        return self.serialize(self.predict(self.deserialize(data)))


class ModelInferenceAPI(LightningWork, abc.ABC):
    def __init__(
        self,
        input: Optional[str] = None,
        output: Optional[str] = None,
        host: str = "127.0.0.1",
        port: int = 7777,
        workers: int = 0,
    ):
        """The ModelInferenceAPI Class enables to easily get your model served.

        Arguments:
            input: Optional `input` to be provided. This would make provide a built-in deserializer.
            output: Optional `output` to be provided. This would make provide a built-in serializer.
            host: Address to be used to serve the model.
            port: Port to be used to serve the model.
            workers: Number of workers for the uvicorn. Warning, this won't work if your subclass takes more arguments.
        """
        super().__init__(parallel=True, host=host, port=port)
        if input and input not in _DESERIALIZER:
            raise Exception(f"Only input in {_DESERIALIZER.keys()} are supported.")
        if output and output not in _SERIALIZER:
            raise Exception(f"Only output in {_SERIALIZER.keys()} are supported.")
        self.input = input
        self.output = output
        self.workers = workers
        self._model = None

        self.ready = False

    @property
    def model(self):
        return self._model

    @abc.abstractmethod
    def build_model(self) -> Any:
        """Override to define your model."""

    def deserialize(self, data) -> Any:
        return data

    @abc.abstractmethod
    def predict(self, data) -> Any:
        """Override to add your predict logic."""

    def serialize(self, data) -> Any:
        return data

    def run(self):
        global fastapi_service
        if self.workers > 1:
            # TODO: This is quite limitated
            # Find a more reliable solution to enable multi workers serving.
            env = os.environ.copy()
            module = inspect.getmodule(self).__file__
            env["LIGHTNING_MODEL_INFERENCE_API_FILE"] = module
            env["LIGHTNING_MODEL_INFERENCE_API_CLASS_NAME"] = self.__class__.__name__
            if self.input:
                env["LIGHTNING_MODEL_INFERENCE_API_INPUT"] = self.input
            if self.output:
                env["LIGHTNING_MODEL_INFERENCE_API_OUTPUT"] = self.output
            command = [
                sys.executable,
                "-m",
                "uvicorn",
                "--workers",
                str(self.workers),
                "--host",
                str(self.host),
                "--port",
                str(self.port),
                "serve:fastapi_service",
            ]
            process = subprocess.Popen(command, env=env, cwd=os.path.dirname(__file__))
            self.ready = True
            process.wait()
        else:
            self._populate_app(fastapi_service)
            self.ready = True
            self._launch_server(fastapi_service)

    def _populate_app(self, fastapi_service: FastAPI):
        self._model = self.build_model()

        fastapi_service.post("/predict", response_class=JSONResponse)(
            _InferenceCallable(
                deserialize=_DESERIALIZER[self.input] if self.input else self.deserialize,
                predict=self.predict,
                serialize=_SERIALIZER[self.output] if self.output else self.serialize,
            ).run
        )

    def _launch_server(self, fastapi_service: FastAPI):
        logger.info(f"Your app has started. View it in your browser: http://{self.host}:{self.port}")
        uvicorn.run(app=fastapi_service, host=self.host, port=self.port, log_level="error")

    def configure_layout(self) -> str:
        return f"{self.url}/docs"


def _maybe_create_instance() -> Optional[ModelInferenceAPI]:
    """This function tries to re-create the user `ModelInferenceAPI` if the environment associated with multi
    workers are present."""
    render_fn_name = os.getenv("LIGHTNING_MODEL_INFERENCE_API_CLASS_NAME", None)
    render_fn_module_file = os.getenv("LIGHTNING_MODEL_INFERENCE_API_FILE", None)
    if render_fn_name is None or render_fn_module_file is None:
        return None
    module = pydoc.importfile(render_fn_module_file)
    cls = getattr(module, render_fn_name)
    input = os.getenv("LIGHTNING_MODEL_INFERENCE_API_INPUT", None)
    output = os.getenv("LIGHTNING_MODEL_INFERENCE_API_OUTPUT", None)
    return cls(input=input, output=output)


instance = _maybe_create_instance()
if instance:
    instance._populate_app(fastapi_service)



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from lightning.app.core.constants import REDIS_HOST, REDIS_PASSWORD, REDIS_PORT
from lightning.app.utilities.imports import _is_redis_available


def check_if_redis_running(
    host: Optional[str] = "", port: Optional[int] = 6379, password: Optional[str] = None
) -> bool:
    if not _is_redis_available():
        return False
    import redis

    try:
        host = host or REDIS_HOST
        port = port or REDIS_PORT
        password = password or REDIS_PASSWORD
        return redis.Redis(host=host, port=port, password=password).ping()
    except redis.exceptions.ConnectionError:
        return False



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for traversing the tree of components in an app."""
from typing import Type, TYPE_CHECKING

import lightning.app

if TYPE_CHECKING:
    from lightning.app.utilities.types import Component, ComponentTuple


def breadth_first(root: "Component", types: Type["ComponentTuple"] = None):
    """Returns a generator that walks through the tree of components breadth-first.

    Arguments:
        root: The root component of the tree
        types: If provided, only the component types in this list will be visited.
    """
    yield from _BreadthFirstVisitor(root, types)


class _BreadthFirstVisitor:
    def __init__(self, root: "Component", types: Type["ComponentTuple"] = None) -> None:
        self.queue = [root]
        self.types = types

    def __iter__(self):
        return self

    def __next__(self):
        from lightning.app.structures import Dict

        while self.queue:
            component = self.queue.pop(0)

            if isinstance(component, lightning.app.LightningFlow):
                components = [getattr(component, el) for el in sorted(component._flows)]
                for struct_name in sorted(component._structures):
                    structure = getattr(component, struct_name)
                    if isinstance(structure, Dict):
                        values = sorted(structure.items(), key=lambda x: x[0])
                    else:
                        values = sorted(((v.name, v) for v in structure), key=lambda x: x[0])
                    for _, value in values:
                        if isinstance(value, lightning.app.LightningFlow):
                            components.append(value)
                self.queue += components
                self.queue += component.works(recurse=False)

            if any(isinstance(component, t) for t in self.types):
                return component

        raise StopIteration


class _DepthFirstVisitor:
    def __init__(self, root: "Component", types: Type["ComponentTuple"] = None) -> None:
        self.stack = [root]
        self.types = types

    def __iter__(self):
        return self

    def __next__(self):
        while self.stack:
            component = self.stack.pop()

            if isinstance(component, lightning.app.LightningFlow):
                self.stack += list(component.flows.values())
                self.stack += component.works(recurse=False)

            if any(isinstance(component, t) for t in self.types):
                return component

        raise StopIteration



import sys
import traceback
from typing import Any, Dict, List, Optional, Tuple

from lightning.app.components.python import TracerPythonScript
from lightning.app.storage import Path
from lightning.app.utilities.packaging.build_config import BuildConfig, load_requirements
from lightning.app.utilities.tracer import Tracer


class ScriptRunner(TracerPythonScript):
    """The ScriptRunner executes the script using ``runpy`` and also patches the Trainer methods to inject
    additional code."""

    def __init__(self, root_path: str, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, cloud_build_config=self._get_build_config(root_path), **kwargs)
        self.root_path = root_path
        self.exception_message: str = ""
        self.trainer_progress: dict = {}
        self.trainer_state: dict = {}
        self.trainer_hparams: dict = {}
        self.model_hparams: dict = {}
        self.log_dir: Optional[Path] = None
        self.checkpoint_dir: Optional[Path] = None
        self.logger_metadatas: List[Dict[str, str]] = []

    def configure_tracer(self) -> Tracer:
        from core.callbacks import PLAppArtifactsTracker, PLAppProgressTracker, PLAppSummary, PLAppTrainerStateTracker

        from lightning.pytorch import Trainer

        tracer = Tracer()
        trainer_artifacts_tracker = PLAppArtifactsTracker(work=self)
        trainer_state_tracker = PLAppTrainerStateTracker(work=self)
        progress_tracker = PLAppProgressTracker(work=self)
        summary = PLAppSummary(work=self)

        def pre_trainer_init(_, *args: Any, **kwargs: Any) -> Tuple[Dict, Tuple[Any, ...], Dict[str, Any]]:
            kwargs.setdefault("callbacks", [])
            kwargs["callbacks"].extend(
                [
                    trainer_artifacts_tracker,
                    trainer_state_tracker,
                    progress_tracker,
                    summary,
                ]
            )
            return {}, args, kwargs

        tracer.add_traced(Trainer, "__init__", pre_fn=pre_trainer_init)
        return tracer

    def run(self) -> None:
        self.exception_message = ""
        # We need to set the module path both in sys.path and the PYTHONPATH env variable.
        # The former is for the current process which is already running, and the env variable is needed in case
        # the script launches subprocesses
        sys.path.insert(0, self.root_path)
        self.env["PYTHONPATH"] = self.root_path
        super().run()

    def on_exception(self, exception: BaseException) -> None:
        self.exception_message = traceback.format_exc()
        super().on_exception(exception)

    @staticmethod
    def _get_build_config(root_path: str) -> Optional[BuildConfig]:
        # These are the requirements for the script runner itself
        requirements = [
            "protobuf<4.21.0",
            "pytorch-lightning<=1.6.3",
            "pydantic<=1.9.0",
        ]
        if Path(root_path, "requirements.txt").exists():
            # Requirements from the user's code folder
            requirements.extend(load_requirements(root_path, file_name="requirements.txt"))

        return BuildConfig(requirements=requirements)



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path
from typing import Optional

import lightning_cloud.env


def get_lightning_cloud_url() -> str:
    # DO NOT CHANGE!
    return os.getenv("LIGHTNING_CLOUD_URL", "https://lightning.ai")


SUPPORTED_PRIMITIVE_TYPES = (type(None), str, int, float, bool)
STATE_UPDATE_TIMEOUT = 0.001
STATE_ACCUMULATE_WAIT = 0.15
# Duration in seconds of a moving average of a full flow execution
# beyond which an exception is raised.
FLOW_DURATION_THRESHOLD = 1.0
# Number of samples for the moving average of the duration of flow execution
FLOW_DURATION_SAMPLES = 5

APP_SERVER_HOST = os.getenv("LIGHTNING_APP_STATE_URL", "http://127.0.0.1")
APP_SERVER_IN_CLOUD = "http://lightningapp" in APP_SERVER_HOST
APP_SERVER_PORT = 7501
APP_STATE_MAX_SIZE_BYTES = 1024 * 1024  # 1 MB

WARNING_QUEUE_SIZE = 1000
# different flag because queue debug can be very noisy, and almost always not useful unless debugging the queue itself.
QUEUE_DEBUG_ENABLED = bool(int(os.getenv("LIGHTNING_QUEUE_DEBUG_ENABLED", "0")))

REDIS_HOST = os.getenv("REDIS_HOST", "localhost")
REDIS_PORT = int(os.getenv("REDIS_PORT", 6379))
REDIS_PASSWORD = os.getenv("REDIS_PASSWORD", None)
REDIS_QUEUES_READ_DEFAULT_TIMEOUT = 0.005

HTTP_QUEUE_URL = os.getenv("LIGHTNING_HTTP_QUEUE_URL", "http://localhost:9801")
HTTP_QUEUE_REFRESH_INTERVAL = float(os.getenv("LIGHTNING_HTTP_QUEUE_REFRESH_INTERVAL", "1"))
HTTP_QUEUE_TOKEN = os.getenv("LIGHTNING_HTTP_QUEUE_TOKEN", None)

USER_ID = os.getenv("USER_ID", "1234")
FRONTEND_DIR = str(Path(__file__).parent.parent / "ui")
PACKAGE_LIGHTNING = os.getenv("PACKAGE_LIGHTNING", None)
CLOUD_UPLOAD_WARNING = int(os.getenv("CLOUD_UPLOAD_WARNING", "2"))
DISABLE_DEPENDENCY_CACHE = bool(int(os.getenv("DISABLE_DEPENDENCY_CACHE", "0")))
# Project under which the resources need to run in cloud. If this env is not set,
# cloud runner will try to get the default project from the cloud
LIGHTNING_CLOUD_PROJECT_ID = os.getenv("LIGHTNING_CLOUD_PROJECT_ID")
LIGHTNING_CLOUD_PRINT_SPECS = os.getenv("LIGHTNING_CLOUD_PRINT_SPECS")
LIGHTNING_DIR = os.getenv("LIGHTNING_DIR", str(Path.home() / ".lightning"))
LIGHTNING_CREDENTIAL_PATH = os.getenv("LIGHTNING_CREDENTIAL_PATH", str(Path(LIGHTNING_DIR) / "credentials.json"))
DOT_IGNORE_FILENAME = ".lightningignore"
LIGHTNING_COMPONENT_PUBLIC_REGISTRY = "https://lightning.ai/v1/components"
LIGHTNING_APPS_PUBLIC_REGISTRY = "https://lightning.ai/v1/apps"

LIGHTNING_CLOUDSPACE_HOST = os.getenv("LIGHTNING_CLOUDSPACE_HOST")
LIGHTNING_CLOUDSPACE_EXPOSED_PORT_COUNT = int(os.getenv("LIGHTNING_CLOUDSPACE_EXPOSED_PORT_COUNT", "0"))

# EXPERIMENTAL: ENV VARIABLES TO ENABLE MULTIPLE WORKS IN THE SAME MACHINE
DEFAULT_NUMBER_OF_EXPOSED_PORTS = int(os.getenv("DEFAULT_NUMBER_OF_EXPOSED_PORTS", "50"))
ENABLE_MULTIPLE_WORKS_IN_NON_DEFAULT_CONTAINER = bool(
    int(os.getenv("ENABLE_MULTIPLE_WORKS_IN_NON_DEFAULT_CONTAINER", "0"))
)  # This isn't used in the cloud yet.

# env var trigger running setup commands in the app
ENABLE_APP_COMMENT_COMMAND_EXECUTION = bool(int(os.getenv("ENABLE_APP_COMMENT_COMMAND_EXECUTION", "0")))


DEBUG: bool = lightning_cloud.env.DEBUG
DEBUG_ENABLED = bool(int(os.getenv("LIGHTNING_DEBUG", "0")))
ENABLE_PULLING_STATE_ENDPOINT = bool(int(os.getenv("ENABLE_PULLING_STATE_ENDPOINT", "1")))
ENABLE_PUSHING_STATE_ENDPOINT = ENABLE_PULLING_STATE_ENDPOINT and bool(
    int(os.getenv("ENABLE_PUSHING_STATE_ENDPOINT", "1"))
)
ENABLE_STATE_WEBSOCKET = bool(int(os.getenv("ENABLE_STATE_WEBSOCKET", "0")))
ENABLE_UPLOAD_ENDPOINT = bool(int(os.getenv("ENABLE_UPLOAD_ENDPOINT", "1")))


def enable_multiple_works_in_default_container() -> bool:
    return bool(int(os.getenv("ENABLE_MULTIPLE_WORKS_IN_DEFAULT_CONTAINER", "0")))


def get_cloud_queue_type() -> Optional[str]:
    value = os.getenv("LIGHTNING_CLOUD_QUEUE_TYPE", None)
    if value is None and enable_interruptible_works():
        value = "http"
    return value


# Number of seconds to wait between filesystem checks when waiting for files in remote storage
REMOTE_STORAGE_WAIT = 0.5


# interruptible support
def enable_interruptible_works() -> bool:
    return bool(int(os.getenv("LIGHTNING_INTERRUPTIBLE_WORKS", "0")))


# Get Cluster Driver
_CLUSTER_DRIVERS = [None, "k8s", "direct"]


def get_cluster_driver() -> Optional[str]:
    value = os.getenv("LIGHTNING_CLUSTER_DRIVER", None)
    if value is None:
        if enable_interruptible_works():
            value = "direct"
        else:
            value = None
    if value not in _CLUSTER_DRIVERS:
        raise ValueError(f"Found {value} cluster driver. The value needs to be in {_CLUSTER_DRIVERS}.")
    return value



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from random import choice, randint

_adjectives = [
    # Appearance, sound, smell...
    "acrid",
    "ambrosial",
    "amorphous",
    "armored",
    "aromatic",
    "bald",
    "blazing",
    "boisterous",
    "bouncy",
    "brawny",
    "bulky",
    "camouflaged",
    "caped",
    "chubby",
    "curvy",
    "elastic",
    "ethereal",
    "fat",
    "feathered",
    "fiery",
    "flashy",
    "flat",
    "fluffy",
    "foamy",
    "fragrant",
    "furry",
    "fuzzy",
    "glaring",
    "hairy",
    "heavy",
    "hissing",
    "horned",
    "icy",
    "imaginary",
    "invisible",
    "lean",
    "loud",
    "loutish",
    "lumpy",
    "lush",
    "masked",
    "meaty",
    "messy",
    "misty",
    "nebulous",
    "noisy",
    "nondescript",
    "organic",
    "purring",
    "quiet",
    "quirky",
    "radiant",
    "roaring",
    "ruddy",
    "rustling",
    "screeching",
    "shaggy",
    "shapeless",
    "shiny",
    "silent",
    "silky",
    "singing",
    "skinny",
    "smooth",
    "soft",
    "spicy",
    "spiked",
    "statuesque",
    "sticky",
    "tacky",
    "tall",
    "tangible",
    "tentacled",
    "thick",
    "thundering",
    "venomous",
    "warm",
    "weightless",
    "whispering",
    "winged",
    "wooden",
    # Beauty & Charm",
    "adorable",
    "affable",
    "amazing",
    "amiable",
    "attractive",
    "beautiful",
    "calm",
    "charming",
    "cherubic",
    "classic",
    "classy",
    "convivial",
    "cordial",
    "cuddly",
    "curly",
    "cute",
    "debonair",
    "elegant",
    "famous",
    "fresh",
    "friendly",
    "funny",
    "gorgeous",
    "graceful",
    "gregarious",
    "grinning",
    "handsome",
    "hilarious",
    "hot",
    "interesting",
    "kind",
    "laughing",
    "lovely",
    "meek",
    "mellow",
    "merciful",
    "neat",
    "nifty",
    "notorious",
    "poetic",
    "pretty",
    "refined",
    "refreshing",
    "sexy",
    "smiling",
    "sociable",
    "spiffy",
    "stylish",
    "sweet",
    "tactful",
    "whimsical",
    "boring",
    # Character & Emotions",
    "abiding",
    "accurate",
    "adamant",
    "adaptable",
    "adventurous",
    "alluring",
    "aloof",
    "ambitious",
    "amusing",
    "annoying",
    "arrogant",
    "aspiring",
    "belligerent",
    "benign",
    "berserk",
    "benevolent",
    "bold",
    "brave",
    "cheerful",
    "chirpy",
    "cocky",
    "congenial",
    "courageous",
    "cryptic",
    "curious",
    "daft",
    "dainty",
    "daring",
    "defiant",
    "delicate",
    "delightful",
    "determined",
    "devout",
    "didactic",
    "diligent",
    "discreet",
    "dramatic",
    "dynamic",
    "eager",
    "eccentric",
    "elated",
    "encouraging",
    "enigmatic",
    "enthusiastic",
    "evasive",
    "faithful",
    "fair",
    "fanatic",
    "fearless",
    "fervent",
    "festive",
    "fierce",
    "fine",
    "free",
    "gabby",
    "garrulous",
    "gay",
    "gentle",
    "glistening",
    "greedy",
    "grumpy",
    "happy",
    "honest",
    "hopeful",
    "hospitable",
    "impetuous",
    "independent",
    "industrious",
    "innocent",
    "intrepid",
    "jolly",
    "jovial",
    "just",
    "lively",
    "loose",
    "loyal",
    "merry",
    "modest",
    "mysterious",
    "nice",
    "obedient",
    "optimistic",
    "orthodox",
    "outgoing",
    "outrageous",
    "overjoyed",
    "passionate",
    "perky",
    "placid",
    "polite",
    "positive",
    "proud",
    "prudent",
    "puzzling",
    "quixotic",
    "quizzical",
    "rebel",
    "resolute",
    "rampant",
    "righteous",
    "romantic",
    "rough",
    "rousing",
    "sassy",
    "satisfied",
    "sly",
    "sincere",
    "snobbish",
    "spirited",
    "spry",
    "stalwart",
    "stirring",
    "swinging",
    "tasteful",
    "thankful",
    "tidy",
    "tremendous",
    "truthful",
    "unselfish",
    "upbeat",
    "uppish",
    "valiant",
    "vehement",
    "vengeful",
    "vigorous",
    "vivacious",
    "zealous",
    "zippy",
    # Intelligence & Abilities",
    "able",
    "adept",
    "analytic",
    "astute",
    "attentive",
    "brainy",
    "busy",
    "calculating",
    "capable",
    "careful",
    "cautious",
    "certain",
    "clever",
    "competent",
    "conscious",
    "cooperative",
    "crafty",
    "crazy",
    "cunning",
    "daffy",
    "devious",
    "discerning",
    "efficient",
    "expert",
    "functional",
    "gifted",
    "helpful",
    "enlightened",
    "idealistic",
    "impartial",
    "industrious",
    "ingenious",
    "inquisitive",
    "intelligent",
    "inventive",
    "judicious",
    "keen",
    "knowing",
    "literate",
    "logical",
    "masterful",
    "mindful",
    "nonchalant",
    "observant",
    "omniscient",
    "poised",
    "practical",
    "pragmatic",
    "proficient",
    "provocative",
    "qualified",
    "radical",
    "rational",
    "realistic",
    "resourceful",
    "savvy",
    "sceptical",
    "sensible",
    "serious",
    "shrewd",
    "skilled",
    "slick",
    "slim",
    "sloppy",
    "smart",
    "sophisticated",
    "stoic",
    "succinct",
    "talented",
    "thoughtful",
    "tricky",
    "unbiased",
    "uptight",
    "versatile",
    "versed",
    "visionary",
    "wise",
    "witty",
    # Strength & Agility",
    "accelerated",
    "active",
    "agile",
    "athletic",
    "dashing",
    "deft",
    "dexterous",
    "energetic",
    "fast",
    "frisky",
    "hasty",
    "hypersonic",
    "meteoric",
    "mighty",
    "muscular",
    "nimble",
    "nippy",
    "powerful",
    "prompt",
    "quick",
    "rapid",
    "resilient",
    "robust",
    "rugged",
    "solid",
    "speedy",
    "steadfast",
    "steady",
    "strong",
    "sturdy",
    "tireless",
    "tough",
    "unyielding",
    # Money & Power",
    "rich",
    "wealthy",
    # Science",
    "meticulous",
    "precise",
    "rigorous",
    "scrupulous",
    "strict",
    # Movement type",
    "airborne",
    "burrowing",
    "crouching",
    "flying",
    "hidden",
    "hopping",
    "jumping",
    "lurking",
    "tunneling",
    "warping",
    # Location and Dwelling",
    "aboriginal",
    "amphibian",
    "aquatic",
    "arboreal",
    "polar",
    "terrestrial",
    "urban",
    # Awesome",
    "accomplished",
    "astonishing",
    "authentic",
    "awesome",
    "delectable",
    "excellent",
    "exotic",
    "exuberant",
    "fabulous",
    "fantastic",
    "fascinating",
    "flawless",
    "fortunate",
    "funky",
    "godlike",
    "glorious",
    "groovy",
    "honored",
    "illustrious",
    "imposing",
    "important",
    "impressive",
    "incredible",
    "invaluable",
    "kickass",
    "majestic",
    "magnificent",
    "marvellous",
    "monumental",
    "perfect",
    "phenomenal",
    "pompous",
    "precious",
    "premium",
    "private",
    "remarkable",
    "spectacular",
    "splendid",
    "successful",
    "wonderful",
    "wondrous",
    # Original",
    "offbeat",
    "original",
    "outstanding",
    "quaint",
    "unique",
    # Time",
    "ancient",
    "antique",
    "prehistoric",
    "primitive",
    # Misc",
    "abstract",
    "acoustic",
    "angelic",
    "arcane",
    "archetypal",
    "augmented",
    "auspicious",
    "axiomatic",
    "beneficial",
    "bipedal",
    "bizarre",
    "complex",
    "dancing",
    "dangerous",
    "demonic",
    "divergent",
    "economic",
    "electric",
    "elite",
    "eminent",
    "enchanted",
    "esoteric",
    "finicky",
    "fractal",
    "futuristic",
    "gainful",
    "hallowed",
    "heavenly",
    "heretic",
    "holistic",
    "hungry",
    "hypnotic",
    "hysterical",
    "illegal",
    "imperial",
    "imported",
    "impossible",
    "inescapable",
    "juicy",
    "liberal",
    "ludicrous",
    "lyrical",
    "magnetic",
    "manipulative",
    "mature",
    "military",
    "macho",
    "married",
    "melodic",
    "natural",
    "naughty",
    "nocturnal",
    "nostalgic",
    "optimal",
    "pastoral",
    "peculiar",
    "piquant",
    "pristine",
    "prophetic",
    "psychedelic",
    "quantum",
    "rare",
    "real",
    "secret",
    "simple",
    "spectral",
    "spiritual",
    "stereotyped",
    "stimulating",
    "straight",
    "strange",
    "tested",
    "therapeutic",
    "true",
    "ubiquitous",
    "uncovered",
    "unnatural",
    "utopian",
    "vagabond",
    "vague",
    "vegan",
    "victorious",
    "vigilant",
    "voracious",
    "wakeful",
    "wandering",
    "watchful",
    "wild",
    # Pseudo-colors",
    "bright",
    "brilliant",
    "colorful",
    "crystal",
    "dark",
    "dazzling",
    "fluorescent",
    "glittering",
    "glossy",
    "gleaming",
    "light",
    "mottled",
    "neon",
    "opalescent",
    "pastel",
    "smoky",
    "sparkling",
    "spotted",
    "striped",
    "translucent",
    "transparent",
    "vivid",
]

# Docker, starting from 0.7.x, generated names from notable scientists and hackers.
# Please, for any amazing man that you add to the list, consider adding an equally amazing woman to it, and vice versa.
_surnames = [
    # Muhammad ibn Jabir Al-Battani was a founding father of astronomy. https://en.wikipedia.org/wiki/Al-Battani
    "albattani",
    # Frances E. Allen, became the first female IBM Fellow in 1989. In 2006, she became the first female
    # recipient of the ACM's Turing Award. https://en.wikipedia.org/wiki/Frances_E._Allen
    "allen",
    # June Almeida - Scottish virologist who took the first pictures of the rubella
    # virus - https://en.wikipedia.org/wiki/June_Almeida
    "almeida",
    # Kathleen Antonelli, American computer programmer and one of the six original
    # programmers of the ENIAC - https://en.wikipedia.org/wiki/Kathleen_Antonelli
    "antonelli",
    # Maria Gaetana Agnesi - Italian mathematician, philosopher, theologian and humanitarian.
    # She was the first woman to write a mathematics handbook and the first woman appointed
    # as a Mathematics Professor at a University. https://en.wikipedia.org/wiki/Maria_Gaetana_Agnesi
    "agnesi",
    # Archimedes was a physicist, engineer and mathematician who invented too many
    # things to list them here. https://en.wikipedia.org/wiki/Archimedes
    "archimedes",
    # Maria Ardinghelli - Italian translator, mathematician and physicist -
    # https://en.wikipedia.org/wiki/Maria_Ardinghelli
    "ardinghelli",
    # Aryabhata - Ancient Indian mathematician-astronomer during 476-550 CE https://en.wikipedia.org/wiki/Aryabhata
    "aryabhata",
    # Wanda Austin - Wanda Austin is the President and CEO of The Aerospace Corporation,
    # a leading architect for the US security space programs. https://en.wikipedia.org/wiki/Wanda_Austin
    "austin",
    # Charles Babbage invented the concept of a programmable computer. https://en.wikipedia.org/wiki/Charles_Babbage.
    "babbage",
    # Stefan Banach - Polish mathematician, was one of the founders of modern
    # functional analysis. https://en.wikipedia.org/wiki/Stefan_Banach
    "banach",
    # Buckaroo Banzai and his mentor Dr. Hikita perfected the "oscillation overthruster",
    # a device that allows one to pass through solid matter. -
    # https://en.wikipedia.org/wiki/The_Adventures_of_Buckaroo_Banzai_Across_the_8th_Dimension
    "banzai",
    # John Bardeen co-invented the transistor - https://en.wikipedia.org/wiki/John_Bardeen
    "bardeen",
    # Jean Bartik, born Betty Jean Jennings, was one of the original programmers
    # for the ENIAC computer. https://en.wikipedia.org/wiki/Jean_Bartik
    "bartik",
    # Laura Bassi, the world's first female professor https://en.wikipedia.org/wiki/Laura_Bassi
    "bassi",
    # Hugh Beaver, British engineer, founder of the Guinness Book of World
    # Records https://en.wikipedia.org/wiki/Hugh_Beaver
    "beaver",
    # Alexander Graham Bell - an eminent Scottish-born scientist, inventor, engineer and innovator
    # who is credited with inventing the first practical telephone - https://en.wikipedia.org/wiki/Alexander_Graham_Bell
    "bell",
    # Karl Friedrich Benz - a German automobile engineer. Inventor of the first
    # practical motorcar. https://en.wikipedia.org/wiki/Karl_Benz
    "benz",
    # Homi J Bhabha - was an Indian nuclear physicist, founding director, and professor of
    # physics at the Tata Institute of Fundamental Research. Colloquially known as "father of
    # Indian nuclear programme"- https://en.wikipedia.org/wiki/Homi_J._Bhabha
    "bhabha",
    # Bhaskara II - Ancient Indian mathematician-astronomer whose work on calculus predates
    # Newton and Leibniz by over half a millennium - https://en.wikipedia.org/wiki/Bh%C4%81skara_II#Calculus
    "bhaskara",
    # Sue Black - British computer scientist and campaigner. She has been instrumental in
    # saving Bletchley Park, the site of World War II codebreaking -
    # https://en.wikipedia.org/wiki/Sue_Black_(computer_scientist)
    "black",
    # Elizabeth Helen Blackburn - Australian-American Nobel laureate; best known
    # for co-discovering telomerase. https://en.wikipedia.org/wiki/Elizabeth_Blackburn
    "blackburn",
    # Elizabeth Blackwell - American doctor and first American woman to receive a
    # medical degree - https://en.wikipedia.org/wiki/Elizabeth_Blackwell
    "blackwell",
    # Niels Bohr is the father of quantum theory. https://en.wikipedia.org/wiki/Niels_Bohr.
    "bohr",
    # Kathleen Booth, she's credited with writing the first assembly language.
    # https://en.wikipedia.org/wiki/Kathleen_Booth
    "booth",
    # Anita Borg - Anita Borg was the founding director of the Institute for
    # Women and Technology (IWT). https://en.wikipedia.org/wiki/Anita_Borg
    "borg",
    # Satyendra Nath Bose - He provided the foundation for Bose\u2013Einstein statistics
    # and the theory of the Bose\u2013Einstein condensate. - https://en.wikipedia.org/wiki/Satyendra_Nath_Bose
    "bose",
    # Katherine Louise Bouman is an imaging scientist and Assistant Professor of Computer
    # Science at the California Institute of Technology. She researches computational methods for
    # imaging, and developed an algorithm that made possible the picture first visualization of a
    # black hole using the Event Horizon Telescope. - https://en.wikipedia.org/wiki/Katie_Bouman
    "bouman",
    # Evelyn Boyd Granville - She was one of the first African-American woman to receive a Ph.D.
    # in mathematics; she earned it in 1949 from Yale University. https://en.wikipedia.org/wiki/Evelyn_Boyd_Granville
    "boyd",
    # Brahmagupta - Ancient Indian mathematician during 598-670 CE who gave rules
    # to compute with zero - https://en.wikipedia.org/wiki/Brahmagupta#Zero
    "brahmagupta",
    # Walter Houser Brattain co-invented the transistor - https://en.wikipedia.org/wiki/Walter_Houser_Brattain
    "brattain",
    # Emmett Brown invented time travel. https://en.wikipedia.org/wiki/Emmett_Brown (thanks Brian Goff)
    "brown",
    # Linda Brown Buck - American biologist and Nobel laureate best known for her genetic and
    # molecular analyses of the mechanisms of smell. https://en.wikipedia.org/wiki/Linda_B._Buck
    "buck",
    # Dame Susan Jocelyn Bell Burnell - Northern Irish astrophysicist who discovered radio pulsars
    # and was the first to analyse them. https://en.wikipedia.org/wiki/Jocelyn_Bell_Burnell
    "burnell",
    # Annie Jump Cannon - pioneering female astronomer who classified hundreds of thousands of stars
    # and created the system we use to understand stars today. https://en.wikipedia.org/wiki/Annie_Jump_Cannon
    "cannon",
    # Rachel Carson - American marine biologist and conservationist, her book Silent Spring and other
    # writings are credited with advancing the global environmental movement.
    # https://en.wikipedia.org/wiki/Rachel_Carson
    "carson",
    # Dame Mary Lucy Cartwright - British mathematician who was one of the first to study what is
    # now known as chaos theory. Also known for Cartwright's theorem which finds applications in
    # signal processing. https://en.wikipedia.org/wiki/Mary_Cartwright
    "cartwright",
    # George Washington Carver - American agricultural scientist and inventor. He was the most
    # prominent black scientist of the early 20th century. https://en.wikipedia.org/wiki/George_Washington_Carver
    "carver",
    # Vinton Gray Cerf - American Internet pioneer, recognised as one of "the fathers of the Internet".
    # With Robert Elliot Kahn, he designed TCP and IP, the primary data communication protocols of
    # the Internet and other computer networks. https://en.wikipedia.org/wiki/Vint_Cerf
    "cerf",
    # Subrahmanyan Chandrasekhar - Astrophysicist known for his mathematical theory on different
    # stages and evolution in structures of the stars. He has won nobel prize for physics -
    # https://en.wikipedia.org/wiki/Subrahmanyan_Chandrasekhar
    "chandrasekhar",
    # Sergey Alexeyevich Chaplygin (April 5, 1869 - October 8, 1942) was a Russian and Soviet physicist,
    # mathematician, and mechanical engineer. He is known for mathematical formulas such as Chaplygin's
    # equation and for a hypothetical substance in cosmology called Chaplygin gas,
    # named after him. https://en.wikipedia.org/wiki/Sergey_Chaplygin
    "chaplygin",
    # Emilie du Chatelet - French natural philosopher, mathematician, physicist, and author
    # during the early 1730s, known for her translation of and commentary on Isaac Newton's book
    # Principia containing basic laws of physics. https://en.wikipedia.org/wiki/%C3%89milie_du_Ch%C3%A2telet
    "chatelet",
    # Asima Chatterjee was an Indian organic chemist noted for her research on vinca alkaloids,
    # development of drugs for treatment of epilepsy and malaria - https://en.wikipedia.org/wiki/Asima_Chatterjee
    "chatterjee",
    # Pafnuty Chebyshev - Russian mathematician. He is known fo his works on probability,
    # statistics, mechanics, analytical geometry and number theory https://en.wikipedia.org/wiki/Pafnuty_Chebyshev
    "chebyshev",
    # Bram Cohen - American computer programmer and author of the BitTorrent
    # peer-to-peer protocol. https://en.wikipedia.org/wiki/Bram_Cohen
    "cohen",
    # David Lee Chaum - American computer scientist and cryptographer. Known for his
    # seminal contributions in the field of anonymous communication. https://en.wikipedia.org/wiki/David_Chaum
    "chaum",
    # Joan Clarke - Bletchley Park code breaker during the Second World War who pioneered techniques
    # that remained top secret for decades. Also an accomplished numismatist https://en.wikipedia.org/wiki/Joan_Clarke
    "clarke",
    # Jane Colden - American botanist widely considered the first female
    # American botanist - https://en.wikipedia.org/wiki/Jane_Colden
    "colden",
    # Gerty Theresa Cori - American biochemist who became the third woman and first American woman to win a
    # Nobel Prize in science, and the first woman to be awarded the Nobel Prize in Physiology
    # or Medicine. Cori was born in Prague. https://en.wikipedia.org/wiki/Gerty_Cori
    "cori",
    # Seymour Roger Cray was an American electrical engineer and supercomputer architect who designed a series
    # of computers that were the fastest in the world for decades. https://en.wikipedia.org/wiki/Seymour_Cray
    "cray",
    # This entry reflects a husband and wife team who worked together:
    # Joan Curran was a Welsh scientist who developed radar and invented chaff, a radar countermeasure.
    # https://en.wikipedia.org/wiki/Joan_Curran Samuel Curran was an Irish physicist who worked
    # alongside his wife during WWII and  invented the proximity fuse. https://en.wikipedia.org/wiki/Samuel_Curran
    "curran",
    # Marie Curie discovered radioactivity. https://en.wikipedia.org/wiki/Marie_Curie.
    "curie",
    # Charles Darwin established the principles of natural evolution. https://en.wikipedia.org/wiki/Charles_Darwin.
    "darwin",
    # Leonardo Da Vinci invented too many things to list here. https://en.wikipedia.org/wiki/Leonardo_da_Vinci.
    "davinci",
    # A. K. (Alexander Keewatin) Dewdney, Canadian mathematician, computer scientist, author and filmmaker.
    # Contributor to Scientific American's "Computer Recreations" from 1984 to 1991. Author of Core War (program),
    # The Planiverse, The Armchair Universe, The Magic Machine, The New Turing Omnibus, and more.
    # https://en.wikipedia.org/wiki/Alexander_Dewdney
    "dewdney",
    # Satish Dhawan - Indian mathematician and aerospace engineer, known for leading the successful and
    # indigenous development of the Indian space programme. https://en.wikipedia.org/wiki/Satish_Dhawan
    "dhawan",
    # Bailey Whitfield Diffie - American cryptographer and one of the pioneers of
    # public-key cryptography. https://en.wikipedia.org/wiki/Whitfield_Diffie
    "diffie",
    # Edsger Wybe Dijkstra was a Dutch computer scientist and mathematical scientist.
    # https://en.wikipedia.org/wiki/Edsger_W._Dijkstra.
    "dijkstra",
    # Paul Adrien Maurice Dirac - English theoretical physicist who made fundamental contributions to the
    # early development of both quantum mechanics and quantum electrodynamics. https://en.wikipedia.org/wiki/Paul_Dirac
    "dirac",
    # Agnes Meyer Driscoll - American cryptanalyst during World Wars I and II who successfully cryptanalysed a
    # number of Japanese ciphers. She was also the co-developer of one of the cipher machines of
    # the US Navy, the CM. https://en.wikipedia.org/wiki/Agnes_Meyer_Driscoll
    "driscoll",
    # Donna Dubinsky - played an integral role in the development of personal digital assistants (PDAs)
    # serving as CEO of Palm, Inc. and co-founding Handspring. https://en.wikipedia.org/wiki/Donna_Dubinsky
    "dubinsky",
    # Annie Easley - She was a leading member of the team which developed software for the Centaur
    # rocket stage and one of the first African-Americans in her field. https://en.wikipedia.org/wiki/Annie_Easley
    "easley",
    # Thomas Alva Edison, prolific inventor https://en.wikipedia.org/wiki/Thomas_Edison
    "edison",
    # Albert Einstein invented the general theory of relativity. https://en.wikipedia.org/wiki/Albert_Einstein
    "einstein",
    # Alexandra Asanovna Elbakyan is a Kazakhstani graduate student, computer programmer, internet pirate in
    # hiding, and the creator of the site Sci-Hub. Nature has listed her in 2016 in the top ten people that
    # mattered in science, and Ars Technica has compared her to Aaron Swartz. -
    # https://en.wikipedia.org/wiki/Alexandra_Elbakyan
    "elbakyan",
    # Taher A. ElGamal - Egyptian cryptographer best known for the ElGamal discrete log cryptosystem and the
    # ElGamal digital signature scheme. https://en.wikipedia.org/wiki/Taher_Elgamal
    "elgamal",
    # Gertrude Elion - American biochemist, pharmacologist and the 1988 recipient of the
    # Nobel Prize in Medicine - https://en.wikipedia.org/wiki/Gertrude_Elion
    "elion",
    # James Henry Ellis - British engineer and cryptographer employed by the GCHQ. Best known for
    # conceiving for the first time, the idea of public-key cryptography. https://en.wikipedia.org/wiki/James_H._Ellis
    "ellis",
    # Douglas Engelbart gave the mother of all demos: https://en.wikipedia.org/wiki/Douglas_Engelbart
    "engelbart",
    # Euclid invented geometry. https://en.wikipedia.org/wiki/Euclid
    "euclid",
    # Leonhard Euler invented large parts of modern mathematics. https://de.wikipedia.org/wiki/Leonhard_Euler
    "euler",
    # Michael Faraday - British scientist who contributed to the study of electromagnetism and
    # electrochemistry. https://en.wikipedia.org/wiki/Michael_Faraday
    "faraday",
    # Horst Feistel - German-born American cryptographer who was one of the earliest non-government
    # researchers to study the design and theory of block ciphers. Co-developer of DES and Lucifer.
    # Feistel networks, a symmetric structure used in the construction of block ciphers are named after him.
    # https://en.wikipedia.org/wiki/Horst_Feistel
    "feistel",
    # Pierre de Fermat pioneered several aspects of modern mathematics. https://en.wikipedia.org/wiki/Pierre_de_Fermat
    "fermat",
    # Enrico Fermi invented the first nuclear reactor. https://en.wikipedia.org/wiki/Enrico_Fermi.
    "fermi",
    # Richard Feynman was a key contributor to quantum mechanics and particle physics.
    # https://en.wikipedia.org/wiki/Richard_Feynman
    "feynman",
    # Benjamin Franklin is famous for his experiments in electricity and the invention of the lightning rod.
    "franklin",
    # Yuri Alekseyevich Gagarin - Soviet pilot and cosmonaut, best known as the first human to
    # journey into outer space. https://en.wikipedia.org/wiki/Yuri_Gagarin
    "gagarin",
    # Galileo was a founding father of modern astronomy, and faced politics and obscurantism to
    # establish scientific truth.  https://en.wikipedia.org/wiki/Galileo_Galilei
    "galileo",
    # Evariste Galois - French mathematician whose work laid the foundations of Galois theory and group theory,
    # two major branches of abstract algebra, and the subfield of Galois connections, all while still in
    # his late teens. https://en.wikipedia.org/wiki/%C3%89variste_Galois
    "galois",
    # Kadambini Ganguly - Indian physician, known for being the first South Asian female physician,
    # trained in western medicine, to graduate in South Asia. https://en.wikipedia.org/wiki/Kadambini_Ganguly
    "ganguly",
    # William Henry "Bill" Gates III is an American business magnate, philanthropist, investor,
    # computer programmer, and inventor. https://en.wikipedia.org/wiki/Bill_Gates
    "gates",
    # Johann Carl Friedrich Gauss - German mathematician who made significant contributions to many fields,
    # including number theory, algebra, statistics, analysis, differential geometry, geodesy, geophysics, mechanics,
    # electrostatics, magnetic fields, astronomy, matrix theory, and optics.
    # https://en.wikipedia.org/wiki/Carl_Friedrich_Gauss
    "gauss",
    # Marie-Sophie Germain - French mathematician, physicist and philosopher. Known for her work o
    # n elasticity theory, number theory and philosophy. https://en.wikipedia.org/wiki/Sophie_Germain
    "germain",
    # Adele Goldberg, was one of the designers and developers of the Smalltalk language.
    # https://en.wikipedia.org/wiki/Adele_Goldberg_(computer_scientist)
    "goldberg",
    # Adele Goldstine, born Adele Katz, wrote the complete technical description for the first electronic
    # digital computer, ENIAC. https://en.wikipedia.org/wiki/Adele_Goldstine
    "goldstine",
    # Shafi Goldwasser is a computer scientist known for creating theoretical foundations of modern
    # cryptography. Winner of 2012 ACM Turing Award. https://en.wikipedia.org/wiki/Shafi_Goldwasser
    "goldwasser",
    # James Golick, all around gangster.
    "golick",
    # Jane Goodall - British primatologist, ethologist, and anthropologist who is considered to be the
    # world's foremost expert on chimpanzees - https://en.wikipedia.org/wiki/Jane_Goodall
    "goodall",
    # Stephen Jay Gould was was an American paleontologist, evolutionary biologist, and historian of science.
    # He is most famous for the theory of punctuated equilibrium - https://en.wikipedia.org/wiki/Stephen_Jay_Gould
    "gould",
    # Carolyn Widney Greider - American molecular biologist and joint winner of the 2009 Nobel Prize for
    # Physiology or Medicine for the discovery of telomerase. https://en.wikipedia.org/wiki/Carol_W._Greider
    "greider",
    # Alexander Grothendieck - German-born French mathematician who became a leading figure in the creation
    # of modern algebraic geometry. https://en.wikipedia.org/wiki/Alexander_Grothendieck
    "grothendieck",
    # Lois Haibt - American computer scientist, part of the team at IBM that developed FORTRAN -
    # https://en.wikipedia.org/wiki/Lois_Haibt
    "haibt",
    # Margaret Hamilton - Director of the Software Engineering Division of the MIT Instrumentation Laboratory,
    # which developed on-board flight software for the Apollo space program.
    # https://en.wikipedia.org/wiki/Margaret_Hamilton_(scientist)
    "hamilton",
    # Caroline Harriet Haslett - English electrical engineer, electricity industry administrator and champion of
    # women's rights. Co-author of British Standard 1363 that specifies AC power plugs and sockets used across
    # the United Kingdom (which is widely considered as one of the safest designs).
    # https://en.wikipedia.org/wiki/Caroline_Haslett
    "haslett",
    # Stephen Hawking pioneered the field of cosmology by combining general relativity and quantum mechanics.
    # https://en.wikipedia.org/wiki/Stephen_Hawking
    "hawking",
    # Martin Edward Hellman - American cryptologist, best known for his invention of public-key cryptography
    # in co-operation with Whitfield Diffie and Ralph Merkle. https://en.wikipedia.org/wiki/Martin_Hellman
    "hellman",
    # Werner Heisenberg was a founding father of quantum mechanics. https://en.wikipedia.org/wiki/Werner_Heisenberg
    "heisenberg",
    # Grete Hermann was a German philosopher noted for her philosophical work on the foundations of quantum mechanics.
    # https://en.wikipedia.org/wiki/Grete_Hermann
    "hermann",
    # Caroline Lucretia Herschel - German astronomer and discoverer of several comets.
    # https://en.wikipedia.org/wiki/Caroline_Herschel
    "herschel",
    # Heinrich Rudolf Hertz - German physicist who first conclusively proved the existence of the electromagnetic waves.
    # https://en.wikipedia.org/wiki/Heinrich_Hertz
    "hertz",
    # Jaroslav Heyrovsky was the inventor of the polarographic method, father of the electroanalytical method, and
    # recipient of the Nobel Prize in 1959. His main field of work was polarography.
    # https://en.wikipedia.org/wiki/Jaroslav_Heyrovsk%C3%BD
    "heyrovsky",
    # Dorothy Hodgkin was a British biochemist, credited with the development of protein crystallography. She was
    # awarded the Nobel Prize in Chemistry in 1964. https://en.wikipedia.org/wiki/Dorothy_Hodgkin
    "hodgkin",
    # Douglas R. Hofstadter is an American professor of cognitive science and author of the Pulitzer Prize and American
    # Book Award-winning work Goedel, Escher, Bach: An Eternal Golden Braid in 1979. A mind-bending work which coined
    # Hofstadter's Law: "It always takes longer than you expect, even when you take into account Hofstadter's Law."
    # https://en.wikipedia.org/wiki/Douglas_Hofstadter
    "hofstadter",
    # Erna Schneider Hoover revolutionized modern communication by inventing a computerized telephone switching method.
    # https://en.wikipedia.org/wiki/Erna_Schneider_Hoover
    "hoover",
    # Grace Hopper developed the first compiler for a computer programming language and  is credited with popularizing
    # the term "debugging" for fixing computer glitches. https://en.wikipedia.org/wiki/Grace_Hopper
    "hopper",
    # Frances Hugle, she was an American scientist, engineer, and inventor who contributed to the understanding of
    # semiconductors, integrated circuitry, and the unique electrical principles of microscopic materials.
    # https://en.wikipedia.org/wiki/Frances_Hugle
    "hugle",
    # Hypatia - Greek Alexandrine Neoplatonist philosopher in Egypt who was one of the earliest mothers of mathematics -
    # https://en.wikipedia.org/wiki/Hypatia
    "hypatia",
    # Teruko Ishizaka - Japanese scientist and immunologist who co-discovered the antibody class Immunoglobulin E.
    # https://en.wikipedia.org/wiki/Teruko_Ishizaka
    "ishizaka",
    # Mary Jackson, American mathematician and aerospace engineer who earned the highest title within NASA's engineering
    #
    # department - https://en.wikipedia.org/wiki/Mary_Jackson_(engineer)
    "jackson",
    # Yeong-Sil Jang was a Korean scientist and astronomer during the Joseon Dynasty; he invented the first metal
    # printing press and water gauge. https://en.wikipedia.org/wiki/Jang_Yeong-sil
    "jang",
    # Mae Carol Jemison -  is an American engineer, physician, and former NASA astronaut. She became the first black
    # woman to travel in space when she served as a mission specialist aboard the Space Shuttle Endeavour -
    # https://en.wikipedia.org/wiki/Mae_Jemison
    "jemison",
    # Betty Jennings - one of the original programmers of the ENIAC. https://en.wikipedia.org/wiki/ENIAC -
    # https://en.wikipedia.org/wiki/Jean_Bartik
    "jennings",
    # Mary Lou Jepsen, was the founder and chief technology officer of One Laptop Per Child (OLPC), and the founder of
    # Pixel Qi. https://en.wikipedia.org/wiki/Mary_Lou_Jepsen
    "jepsen",
    # Katherine Coleman Goble Johnson - American physicist and mathematician contributed to the NASA.
    # https://en.wikipedia.org/wiki/Katherine_Johnson
    "johnson",
    # Irene Joliot-Curie - French scientist who was awarded the Nobel Prize for Chemistry in 1935. Daughter of Marie
    # and Pierre Curie. https://en.wikipedia.org/wiki/Ir%C3%A8ne_Joliot-Curie
    "joliot",
    # Karen Sparck Jones came up with the concept of inverse document frequency, which is used in most search engines
    # today. https://en.wikipedia.org/wiki/Karen_Sp%C3%A4rck_Jones
    "jones",
    # A. P. J. Abdul Kalam - is an Indian scientist aka Missile Man of India for his work on the development of
    # ballistic missile and launch vehicle technology - https://en.wikipedia.org/wiki/A._P._J._Abdul_Kalam
    "kalam",
    # Sergey Petrovich Kapitsa (14 February 1928 - 14 August 2012) was a Russian physicist and demographer. He was best
    # known as host of the popular and long-running Russian scientific TV show, Evident, but Incredible. His father was
    # the Nobel laureate Soviet-era physicist Pyotr Kapitsa, and his brother was the geographer and Antarctic explorer
    # Andrey Kapitsa. - https://en.wikipedia.org/wiki/Sergey_Kapitsa
    "kapitsa",
    # Susan Kare, created the icons and many of the interface elements for the original Apple Macintosh in the 1980s,
    # and was an original employee of NeXT, working as the Creative Director. https://en.wikipedia.org/wiki/Susan_Kare
    "kare",
    # Mstislav Keldysh - a Soviet scientist in the field of mathematics and mechanics, academician of the USSR Academy
    # of Sciences (1946), President of the USSR Academy of Sciences (1961-1975),
    # three times Hero of Socialist Labor (1956, 1961, 1971), fellow of the Royal Society of Edinburgh (1968).
    # https://en.wikipedia.org/wiki/Mstislav_Keldysh
    "keldysh",
    # Mary Kenneth Keller, Sister Mary Kenneth Keller became the first American woman to earn a
    # PhD in Computer Science in 1965. https://en.wikipedia.org/wiki/Mary_Kenneth_Keller
    "keller",
    # Johannes Kepler, German astronomer known for his three laws of planetary motion -
    # https://en.wikipedia.org/wiki/Johannes_Kepler
    "kepler",
    # Omar Khayyam - Persian mathematician, astronomer and poet. Known for his work on the classification and solution
    # of cubic equations, for his contribution to the understanding of Euclid's fifth postulate and for computing the
    # length of a year very accurately. https://en.wikipedia.org/wiki/Omar_Khayyam
    "khayyam",
    # Har Gobind Khorana - Indian-American biochemist who shared the 1968 Nobel Prize for Physiology -
    # https://en.wikipedia.org/wiki/Har_Gobind_Khorana
    "khorana",
    # Jack Kilby invented silicon integrated circuits and gave Silicon Valley its name. -
    # https://en.wikipedia.org/wiki/Jack_Kilby
    "kilby",
    # Maria Kirch - German astronomer and first woman to discover a comet -
    # https://en.wikipedia.org/wiki/Maria_Margarethe_Kirch
    "kirch",
    # Donald Knuth - American computer scientist, author of "The Art of Computer Programming" and creator of the TeX
    # typesetting system. https://en.wikipedia.org/wiki/Donald_Knuth
    "knuth",
    # Sophie Kowalevski - Russian mathematician responsible for important original contributions to analysis,
    # differential equations and mechanics - https://en.wikipedia.org/wiki/Sofia_Kovalevskaya
    "kowalevski",
    # Marie-Jeanne de Lalande - French astronomer, mathematician and cataloguer of stars -
    # https://en.wikipedia.org/wiki/Marie-Jeanne_de_Lalande
    "lalande",
    # Hedy Lamarr - Actress and inventor. The principles of her work are now incorporated into modern Wi-Fi, CDMA
    # and Bluetooth technology. https://en.wikipedia.org/wiki/Hedy_Lamarr
    "lamarr",
    # Leslie B. Lamport - American computer scientist. Lamport is best known for his seminal work in distributed
    # systems and was the winner of the 2013 Turing Award. https://en.wikipedia.org/wiki/Leslie_Lamport
    "lamport",
    # Mary Leakey - British paleoanthropologist who discovered the first fossilized Proconsul skull -
    # https://en.wikipedia.org/wiki/Mary_Leakey
    "leakey",
    # Henrietta Swan Leavitt - she was an American astronomer who discovered the relation between the luminosity and
    # the period of Cepheid variable stars. https://en.wikipedia.org/wiki/Henrietta_Swan_Leavitt
    "leavitt",
    # Esther Miriam Zimmer Lederberg - American microbiologist and a pioneer of bacterial genetics.
    # https://en.wikipedia.org/wiki/Esther_Lederberg
    "lederberg",
    # Inge Lehmann - Danish seismologist and geophysicist. Known for discovering in 1936 that the Earth has a solid
    # inner core inside a molten outer core. https://en.wikipedia.org/wiki/Inge_Lehmann
    "lehmann",
    # Daniel Lewin - Mathematician, Akamai co-founder, soldier, 9/11 victim-- Developed optimization techniques for
    # routing traffic on the internet. Died attempting to stop the 9-11 hijackers.
    # https://en.wikipedia.org/wiki/Daniel_Lewin
    "lewin",
    # Ruth Lichterman - one of the original programmers of the ENIAC. https://en.wikipedia.org/wiki/ENIAC -
    # https://en.wikipedia.org/wiki/Ruth_Teitelbaum
    "lichterman",
    # Barbara Liskov - co-developed the Liskov substitution principle. Liskov was also the winner of the Turing
    # Prize in 2008. - https://en.wikipedia.org/wiki/Barbara_Liskov
    "liskov",
    # Ada Lovelace invented the first algorithm. https://en.wikipedia.org/wiki/Ada_Lovelace (thanks James Turnbull)
    "lovelace",
    # Auguste and Louis Lumiere - the first filmmakers in history -
    # https://en.wikipedia.org/wiki/Auguste_and_Louis_Lumi%C3%A8re
    "lumiere",
    # Mahavira - Ancient Indian mathematician during 9th century AD who discovered basic algebraic identities -
    # https://en.wikipedia.org/wiki/Mah%C4%81v%C4%ABra_(mathematician)
    "mahavira",
    # Lynn Margulis (b. Lynn Petra Alexander) - an American evolutionary theorist and biologist, science author,
    # educator, and popularizer, and was the primary modern proponent for the significance of symbiosis in evolution. -
    # https://en.wikipedia.org/wiki/Lynn_Margulis
    "margulis",
    # Yukihiro Matsumoto - Japanese computer scientist and software programmer best known as the chief designer of
    # the Ruby programming language. https://en.wikipedia.org/wiki/Yukihiro_Matsumoto
    "matsumoto",
    # James Clerk Maxwell - Scottish physicist, best known for his formulation of electromagnetic theory.
    # https://en.wikipedia.org/wiki/James_Clerk_Maxwell
    "maxwell",
    # Maria Mayer - American theoretical physicist and Nobel laureate in Physics for proposing the nuclear shell model
    # of the atomic nucleus - https://en.wikipedia.org/wiki/Maria_Mayer
    "mayer",
    # John McCarthy invented LISP: https://en.wikipedia.org/wiki/John_McCarthy_(computer_scientist)
    "mccarthy",
    # Barbara McClintock - a distinguished American cytogeneticist, 1983 Nobel Laureate in Physiology or Medicine for
    # discovering transposons. https://en.wikipedia.org/wiki/Barbara_McClintock
    "mcclintock",
    # Anne Laura Dorinthea McLaren - British developmental biologist whose work helped lead to human
    # in-vitro fertilisation. https://en.wikipedia.org/wiki/Anne_McLaren
    "mclaren",
    # Malcolm McLean invented the modern shipping container: https://en.wikipedia.org/wiki/Malcom_McLean
    "mclean",
    # Kay McNulty - one of the original programmers of the ENIAC. https://en.wikipedia.org/wiki/ENIAC -
    # https://en.wikipedia.org/wiki/Kathleen_Antonelli
    "mcnulty",
    # Gregor Johann Mendel - Czech scientist and founder of genetics. https://en.wikipedia.org/wiki/Gregor_Mendel
    "mendel",
    # Dmitri Mendeleev - a chemist and inventor. He formulated the Periodic Law, created a farsighted version of the
    # periodic table of elements, and used it to correct the properties of some already discovered elements and also
    # to predict the properties of eight elements yet to be discovered. https://en.wikipedia.org/wiki/Dmitri_Mendeleev
    "mendeleev",
    # Lise Meitner - Austrian/Swedish physicist who was involved in the discovery of nuclear fission. The element
    # meitnerium is named after her - https://en.wikipedia.org/wiki/Lise_Meitner
    "meitner",
    # Carla Meninsky, was the game designer and programmer for Atari 2600 games Dodge 'Em and Warlords.
    # https://en.wikipedia.org/wiki/Carla_Meninsky
    "meninsky",
    # Ralph C. Merkle - American computer scientist, known for devising Merkle's puzzles - one of the very first
    # schemes for public-key cryptography. Also, inventor of Merkle trees and co-inventor of the Merkle-Damgard
    # construction for building collision-resistant cryptographic hash functions and the Merkle-Hellman knapsack
    # cryptosystem. https://en.wikipedia.org/wiki/Ralph_Merkle
    "merkle",
    # Johanna Mestorf - German prehistoric archaeologist and first female museum director in Germany -
    # https://en.wikipedia.org/wiki/Johanna_Mestorf
    "mestorf",
    # Maryam Mirzakhani - an Iranian mathematician and the first woman to win the Fields Medal.
    # https://en.wikipedia.org/wiki/Maryam_Mirzakhani
    "mirzakhani",
    # Rita Levi-Montalcini - Won Nobel Prize in Physiology or Medicine jointly with colleague Stanley Cohen for the
    # discovery of nerve growth factor (https://en.wikipedia.org/wiki/Rita_Levi-Montalcini)
    "montalcini",
    # Gordon Earle Moore - American engineer, Silicon Valley founding father, author of Moore's law.
    # https://en.wikipedia.org/wiki/Gordon_Moore
    "moore",
    # Samuel Morse - contributed to the invention of a single-wire telegraph system based on European telegraphs
    # and was a co-developer of the Morse code - https://en.wikipedia.org/wiki/Samuel_Morse
    "morse",
    # Ian Murdock - founder of the Debian project - https://en.wikipedia.org/wiki/Ian_Murdock
    "murdock",
    # May-Britt Moser - Nobel prize winner neuroscientist who contributed to the discovery of grid cells in the brain.
    # https://en.wikipedia.org/wiki/May-Britt_Moser
    "moser",
    # John Napier of Merchiston - Scottish landowner known as an astronomer, mathematician and physicist.
    # Best known for his discovery of logarithms. https://en.wikipedia.org/wiki/John_Napier
    "napier",
    # John Forbes Nash, Jr. - American mathematician who made fundamental contributions to game theory, differential
    # geometry, and the study of partial differential equations. https://en.wikipedia.org/wiki/John_Forbes_Nash_Jr.
    "nash",
    # John von Neumann - todays computer architectures are based on the von Neumann architecture.
    # https://en.wikipedia.org/wiki/Von_Neumann_architecture
    "neumann",
    # Isaac Newton invented classic mechanics and modern optics. https://en.wikipedia.org/wiki/Isaac_Newton
    "newton",
    # Florence Nightingale, more prominently known as a nurse, was also the first female member of the Royal Statistical
    # Society and a pioneer in statistical graphics
    # https://en.wikipedia.org/wiki/Florence_Nightingale#Statistics_and_sanitary_reform
    "nightingale",
    # Alfred Nobel - a Swedish chemist, engineer, innovator, and armaments manufacturer (inventor of dynamite) -
    # https://en.wikipedia.org/wiki/Alfred_Nobel
    "nobel",
    # Emmy Noether, German mathematician. Noether's Theorem is named after her.
    # https://en.wikipedia.org/wiki/Emmy_Noether
    "noether",
    # Poppy Northcutt. Poppy Northcutt was the first woman to work as part of NASA's Mission Control.
    # http://www.businessinsider.com/poppy-northcutt-helped-apollo-astronauts-2014-12?op=1
    "northcutt",
    # Robert Noyce invented silicon integrated circuits and gave Silicon Valley its name. -
    # https://en.wikipedia.org/wiki/Robert_Noyce
    "noyce",
    # Panini - Ancient Indian linguist and grammarian from 4th century CE who worked on the world's first formal system
    # - https://en.wikipedia.org/wiki/P%C4%81%E1%B9%87ini#Comparison_with_modern_formal_systems
    "panini",
    # Ambroise Pare invented modern surgery. https://en.wikipedia.org/wiki/Ambroise_Par%C3%A9
    "pare",
    # Blaise Pascal, French mathematician, physicist, and inventor - https://en.wikipedia.org/wiki/Blaise_Pascal
    "pascal",
    # Louis Pasteur discovered vaccination, fermentation and pasteurization.
    # https://en.wikipedia.org/wiki/Louis_Pasteur.
    "pasteur",
    # Cecilia Payne-Gaposchkin was an astronomer and astrophysicist who, in 1925, proposed in her Ph.D. thesis an
    # explanation for the composition of stars in terms of the relative abundances of hydrogen and helium.
    # https://en.wikipedia.org/wiki/Cecilia_Payne-Gaposchkin
    "payne",
    # Radia Perlman is a software designer and network engineer and most famous for her invention of the
    # spanning-tree protocol (STP). https://en.wikipedia.org/wiki/Radia_Perlman
    "perlman",
    # Rob Pike was a key contributor to Unix, Plan 9, the X graphic system, utf-8, and the Go programming language.
    # https://en.wikipedia.org/wiki/Rob_Pike
    "pike",
    # Henri Poincare made fundamental contributions in several fields of mathematics.
    # https://en.wikipedia.org/wiki/Henri_Poincar%C3%A9
    "poincare",
    # Laura Poitras is a director and producer whose work, made possible by open source crypto tools, advances the
    # causes of truth and freedom of information by reporting disclosures by whistleblowers such as Edward Snowden.
    # https://en.wikipedia.org/wiki/Laura_Poitras
    "poitras",
    # Tat'yana Avenirovna Proskuriakova (January 23 [O.S. January 10] 1909 - August 30, 1985) was a Russian-American
    # Mayanist scholar and archaeologist who contributed significantly to the deciphering of Maya hieroglyphs, the
    # writing system of the pre-Columbian Maya civilization of Mesoamerica.
    # https://en.wikipedia.org/wiki/Tatiana_Proskouriakoff
    "proskuriakova",
    # Claudius Ptolemy - a Greco-Egyptian writer of Alexandria, known as a mathematician, astronomer, geographer,
    # astrologer, and poet of a single epigram in the Greek Anthology - https://en.wikipedia.org/wiki/Ptolemy
    "ptolemy",
    # C. V. Raman - Indian physicist who won the Nobel Prize in 1930 for proposing the Raman effect. -
    # https://en.wikipedia.org/wiki/C._V._Raman
    "raman",
    # Srinivasa Ramanujan - Indian mathematician and autodidact who made extraordinary contributions to mathematical
    # analysis, number theory, infinite series, and continued fractions. -
    # https://en.wikipedia.org/wiki/Srinivasa_Ramanujan
    "ramanujan",
    # Sally Kristen Ride was an American physicist and astronaut. She was the first American woman in space, and the
    # youngest American astronaut. https://en.wikipedia.org/wiki/Sally_Ride
    "ride",
    # Dennis Ritchie - co-creator of UNIX and the C programming language. - https://en.wikipedia.org/wiki/Dennis_Ritchie
    "ritchie",
    # Ida Rhodes - American pioneer in computer programming, designed the first computer used for Social Security.
    # https://en.wikipedia.org/wiki/Ida_Rhodes
    "rhodes",
    # Julia Hall Bowman Robinson - American mathematician renowned for her contributions to the fields of computability
    # theory and computational complexity theory. https://en.wikipedia.org/wiki/Julia_Robinson
    "robinson",
    # Wilhelm Conrad Rontgen - German physicist who was awarded the first Nobel Prize in Physics in 1901 for the
    # discovery of X-rays (Rontgen rays). https://en.wikipedia.org/wiki/Wilhelm_R%C3%B6ntgen
    "roentgen",
    # Rosalind Franklin - British biophysicist and X-ray crystallographer whose research was critical to the
    # understanding of DNA - https://en.wikipedia.org/wiki/Rosalind_Franklin
    "rosalind",
    # Vera Rubin - American astronomer who pioneered work on galaxy rotation rates.
    # https://en.wikipedia.org/wiki/Vera_Rubin
    "rubin",
    # Meghnad Saha - Indian astrophysicist best known for his development of the Saha equation, used to describe
    # chemical and physical conditions in stars - https://en.wikipedia.org/wiki/Meghnad_Saha
    "saha",
    # Jean E. Sammet developed FORMAC, the first widely used computer language for symbolic manipulation of
    # mathematical formulas. https://en.wikipedia.org/wiki/Jean_E._Sammet
    "sammet",
    # Mildred Sanderson - American mathematician best known for Sanderson's theorem concerning modular invariants.
    # https://en.wikipedia.org/wiki/Mildred_Sanderson
    "sanderson",
    # Satoshi Nakamoto is the name used by the unknown person or group of people who developed bitcoin, authored the
    # bitcoin white paper, and created and deployed bitcoin's original reference implementation.
    # https://en.wikipedia.org/wiki/Satoshi_Nakamoto
    "satoshi",
    # Adi Shamir - Israeli cryptographer whose numerous inventions and contributions to cryptography include the Ferge
    # Fiat Shamir identification scheme, the Rivest Shamir Adleman (RSA) public-key cryptosystem, the Shamir's secret
    # sharing scheme, the breaking of the Merkle-Hellman cryptosystem, the TWINKLE and TWIRL factoring devices and the
    # discovery of differential cryptanalysis (with Eli Biham). https://en.wikipedia.org/wiki/Adi_Shamir
    "shamir",
    # Claude Shannon - The father of information theory and founder of digital circuit design theory.
    # (https://en.wikipedia.org/wiki/Claude_Shannon)
    "shannon",
    # Carol Shaw - Originally an Atari employee, Carol Shaw is said to be the first female video game designer.
    # https://en.wikipedia.org/wiki/Carol_Shaw_(video_game_designer)
    "shaw",
    # Dame Stephanie "Steve" Shirley - Founded a software company in 1962 employing women working from home.
    # https://en.wikipedia.org/wiki/Steve_Shirley
    "shirley",
    # William Shockley co-invented the transistor - https://en.wikipedia.org/wiki/William_Shockley
    "shockley",
    # Lina Solomonovna Stern (or Shtern; 26 August 1878 - 7 March 1968) was a Soviet biochemist, physiologist and
    # humanist whose medical discoveries saved thousands of lives at the fronts of World War II. She is best known
    # for her pioneering work on blood\u2013brain barrier, which she described as hemato-encephalic barrier in 1921.
    # https://en.wikipedia.org/wiki/Lina_Stern
    "shtern",
    # Francoise Barre-Sinoussi - French virologist and Nobel Prize Laureate in Physiology or Medicine; her work was
    # fundamental in identifying HIV as the cause of AIDS.
    # https://en.wikipedia.org/wiki/Fran%C3%A7oise_Barr%C3%A9-Sinoussi
    "sinoussi",
    # Betty Snyder - one of the original programmers of the ENIAC. https://en.wikipedia.org/wiki/ENIAC -
    # https://en.wikipedia.org/wiki/Betty_Holberton
    "snyder",
    # Cynthia Solomon - Pioneer in the fields of artificial intelligence, computer science and educational computing.
    # Known for creation of Logo, an educational programming language.  https://en.wikipedia.org/wiki/Cynthia_Solomon
    "solomon",
    # Frances Spence - one of the original programmers of the ENIAC. https://en.wikipedia.org/wiki/ENIAC -
    # https://en.wikipedia.org/wiki/Frances_Spence
    "spence",
    # Michael Stonebraker is a database research pioneer and architect of Ingres, Postgres, VoltDB and SciDB.
    # Winner of 2014 ACM Turing Award. https://en.wikipedia.org/wiki/Michael_Stonebraker
    "stonebraker",
    # Ivan Edward Sutherland - American computer scientist and Internet pioneer, widely regarded as the father of
    # computer graphics. https://en.wikipedia.org/wiki/Ivan_Sutherland
    "sutherland",
    # Janese Swanson (with others) developed the first of the Carmen Sandiego games. She went on to found Girl Tech.
    # https://en.wikipedia.org/wiki/Janese_Swanson
    "swanson",
    # Aaron Swartz was influential in creating RSS, Markdown, Creative Commons, Reddit, and much of the internet as we
    # know it today. He was devoted to freedom of information on the web. https://en.wikiquote.org/wiki/Aaron_Swartz
    "swartz",
    # Bertha Swirles was a theoretical physicist who made a number of contributions to early quantum theory.
    # https://en.wikipedia.org/wiki/Bertha_Swirles
    "swirles",
    # Helen Brooke Taussig - American cardiologist and founder of the field of paediatric cardiology.
    # https://en.wikipedia.org/wiki/Helen_B._Taussig
    "taussig",
    # Valentina Tereshkova is a Russian engineer, cosmonaut and politician. She was the first woman to fly to space in
    # 1963. In 2013, at the age of 76, she offered to go on a one-way mission to Mars.
    # https://en.wikipedia.org/wiki/Valentina_Tereshkova
    "tereshkova",
    # Nikola Tesla invented the AC electric system and every gadget ever used by a James Bond villain.
    # https://en.wikipedia.org/wiki/Nikola_Tesla
    "tesla",
    # Marie Tharp - American geologist and oceanic cartographer who co-created the first scientific map of the Atlantic
    # Ocean floor. Her work led to the acceptance of the theories of plate tectonics and continental drift.
    # https://en.wikipedia.org/wiki/Marie_Tharp
    "tharp",
    # Ken Thompson - co-creator of UNIX and the C programming language - https://en.wikipedia.org/wiki/Ken_Thompson
    "thompson",
    # Linus Torvalds invented Linux and Git. https://en.wikipedia.org/wiki/Linus_Torvalds
    "torvalds",
    # Youyou Tu - Chinese pharmaceutical chemist and educator known for discovering artemisinin and dihydroartemisinin,
    # used to treat malaria, which has saved millions of lives. Joint winner of the 2015 Nobel Prize in Physiology or
    # Medicine. https://en.wikipedia.org/wiki/Tu_Youyou
    "tu",
    # Alan Turing was a founding father of computer science. https://en.wikipedia.org/wiki/Alan_Turing.
    "turing",
    # Varahamihira - Ancient Indian mathematician who discovered trigonometric formulae during 505-587 CE -
    # https://en.wikipedia.org/wiki/Var%C4%81hamihira#Contributions
    "varahamihira",
    # Dorothy Vaughan was a NASA mathematician and computer programmer on the SCOUT launch vehicle program that put
    # America's first satellites into space - https://en.wikipedia.org/wiki/Dorothy_Vaughan
    "vaughan",
    # Sir Mokshagundam Visvesvaraya - is a notable Indian engineer.  He is a recipient of the Indian Republic's highest
    # honour, the Bharat Ratna, in 1955. On his birthday, 15 September is celebrated as Engineer's Day in India in his
    # memory - https://en.wikipedia.org/wiki/Visvesvaraya
    "visvesvaraya",
    # Christiane Nusslein-Volhard - German biologist, won Nobel Prize in Physiology or Medicine in 1995 for research on
    # the genetic control of embryonic development. https://en.wikipedia.org/wiki/Christiane_N%C3%BCsslein-Volhard
    "volhard",
    # Cedric Villani - French mathematician, won Fields Medal, Fermat Prize and Poincare Price for his work in
    # differential geometry and statistical mechanics. https://en.wikipedia.org/wiki/C%C3%A9dric_Villani
    "villani",
    # Marlyn Wescoff - one of the original programmers of the ENIAC. https://en.wikipedia.org/wiki/ENIAC -
    # https://en.wikipedia.org/wiki/Marlyn_Meltzer
    "wescoff",
    # Sylvia B. Wilbur - British computer scientist who helped develop the ARPANET, was one of the first to exchange
    # email in the UK and a leading researcher in computer-supported collaborative work.
    # https://en.wikipedia.org/wiki/Sylvia_Wilbur
    "wilbur",
    # Andrew Wiles - Notable British mathematician who proved the enigmatic Fermat's Last Theorem -
    # https://en.wikipedia.org/wiki/Andrew_Wiles
    "wiles",
    # Roberta Williams, did pioneering work in graphical adventure games for personal computers, particularly the King's
    # Quest series. https://en.wikipedia.org/wiki/Roberta_Williams
    "williams",
    # Malcolm John Williamson - British mathematician and cryptographer employed by the GCHQ. Developed in 1974 what
    # is now known as Diffie-Hellman key exchange (Diffie and Hellman first published the scheme in 1976).
    # https://en.wikipedia.org/wiki/Malcolm_J._Williamson
    "williamson",
    # Sophie Wilson designed the first Acorn Micro-Computer and the instruction set for ARM processors.
    # https://en.wikipedia.org/wiki/Sophie_Wilson
    "wilson",
    # Jeannette Wing - co-developed the Liskov substitution principle. - https://en.wikipedia.org/wiki/Jeannette_Wing
    "wing",
    # Steve Wozniak invented the Apple I and Apple II. https://en.wikipedia.org/wiki/Steve_Wozniak
    "wozniak",
    # The Wright brothers, Orville and Wilbur - credited with inventing and building the world's first successful
    # airplane and making the first controlled, powered and sustained heavier-than-air human flight -
    # https://en.wikipedia.org/wiki/Wright_brothers
    "wright",
    # Chien-Shiung Wu - Chinese-American experimental physicist who made significant contributions to nuclear physics.
    # https://en.wikipedia.org/wiki/Chien-Shiung_Wu
    "wu",
    # Rosalyn Sussman Yalow - Rosalyn Sussman Yalow was an American medical physicist, and a co-winner of the 1977
    # Nobel Prize in Physiology or Medicine for development of the radioimmunoassay technique.
    # https://en.wikipedia.org/wiki/Rosalyn_Sussman_Yalow
    "yalow",
    # Ada Yonath - an Israeli crystallographer, the first woman from the Middle East to win a Nobel prize in the
    # sciences. https://en.wikipedia.org/wiki/Ada_Yonath
    "yonath",
    # Nikolay Yegorovich Zhukovsky (January 17 1847 - March 17, 1921) was a Russian scientist, mathematician and
    # engineer, and a founding father of modern aero- and hydrodynamics. Whereas contemporary scientists scoffed at the
    # idea of human flight, Zhukovsky was the first to undertake the study of airflow. He is often called the Father
    # of Russian Aviation. https://en.wikipedia.org/wiki/Nikolay_Yegorovich_Zhukovsky
    "zhukovsky",
]


def get_unique_name():
    """Generates a random name in the style of "docker containers".

    This is generated from the list of adjectives and surnames in this package,
    formatted as "adjective_surname" with a random integer between 0 and 10000
    added to the end.

    A python port of docker's random container name generator.
    Original source:
        https://raw.githubusercontent.com/moby/moby/master/pkg/namesgenerator/names-generator.go

    Examples:

    >>> import random ; random.seed(42)
    >>> get_unique_name()
    'meek-ardinghelli-4506'
    >>> get_unique_name()
    'truthful-dijkstra-2286'
    """
    adjective, surname, i = choice(_adjectives), choice(_surnames), randint(0, 9999)
    return f"{adjective}-{surname}-{i}"



import os
from typing import Dict, List, Optional, TYPE_CHECKING

from lightning.app import LightningFlow

if TYPE_CHECKING:
    import wandb


class WeightsAndBiases(LightningFlow):
    def __init__(self, username: str, project_name: str, run_id: str, api_key: Optional[str] = None) -> None:
        super().__init__()
        self.username = username
        self.project_name = project_name
        self.run_id = run_id
        self._api_key = api_key
        self._run: Optional[wandb.Run] = None

    def run(self) -> None:
        if self._run is not None:
            return

        if self._api_key:
            os.environ["WANDB_API_KEY"] = self._api_key

        import wandb

        self._run = wandb.init(project=self.project_name, id=self.run_id, entity=self.username)

    def configure_layout(self) -> List[Dict[str, str]]:
        if self._run is not None:
            return [{"name": "Training Logs", "content": self._run.get_url()}]
        return []



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import inspect
import os
import pydoc
import subprocess
import sys
from typing import Any, Callable, Type

from lightning.app.core.work import LightningWork
from lightning.app.utilities.app_helpers import StreamLitStatePlugin
from lightning.app.utilities.state import AppState


class ServeStreamlit(LightningWork, abc.ABC):
    """The ``ServeStreamlit`` work allows you to use streamlit from a work.

    You can optionally build a model in the ``build_model`` hook, which will only be called once per session.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.ready = False

        self._process = None

    @property
    def model(self) -> Any:
        return getattr(self, "_model", None)

    @abc.abstractmethod
    def render(self) -> None:
        """Override with your streamlit render function."""

    def build_model(self) -> Any:
        """Optionally override to instantiate and return your model.

        The model will be accessible under ``self.model``.
        """
        return None

    def run(self) -> None:
        env = os.environ.copy()
        env["LIGHTNING_COMPONENT_NAME"] = self.name
        env["LIGHTNING_WORK"] = self.__class__.__name__
        env["LIGHTNING_WORK_MODULE_FILE"] = inspect.getmodule(self).__file__
        self._process = subprocess.Popen(
            [
                sys.executable,
                "-m",
                "streamlit",
                "run",
                __file__,
                "--server.address",
                str(self.host),
                "--server.port",
                str(self.port),
                "--server.headless",
                "true",  # do not open the browser window when running locally
            ],
            env=env,
        )
        self.ready = True
        self._process.wait()

    def on_exit(self) -> None:
        if self._process is not None:
            self._process.kill()

    def configure_layout(self) -> str:
        return self.url


class _PatchedWork:
    """The ``_PatchedWork`` is used to emulate a work instance from a subprocess. This is acheived by patching the
    self reference in methods an properties to point to the AppState.

    Args:
        state: The work state to patch
        work_class: The work class to emulate
    """

    def __init__(self, state: AppState, work_class: Type):
        super().__init__()
        self._state = state
        self._work_class = work_class

    def __getattr__(self, name: str) -> Any:
        try:
            return getattr(self._state, name)
        except AttributeError:
            # The name isn't in the state, so check if it's a callable or a property
            attribute = inspect.getattr_static(self._work_class, name)
            if callable(attribute):
                attribute = attribute.__get__(self, self._work_class)
                return attribute
            elif isinstance(attribute, (staticmethod, property)):
                return attribute.__get__(self, self._work_class)

            # Look for the name in the instance (e.g. for private variables)
            return object.__getattribute__(self, name)

    def __setattr__(self, name: str, value: Any) -> None:
        if name in ["_state", "_work_class"]:
            return object.__setattr__(self, name, value)

        if hasattr(self._state, name):
            return setattr(self._state, name, value)
        return object.__setattr__(self, name, value)


def _reduce_to_component_scope(state: AppState, component_name: str) -> AppState:
    """Given the app state, this utility traverses down to the level of the given component name."""
    component_name_parts = component_name.split(".")[1:]  # exclude root
    component_state = state
    for part in component_name_parts:
        component_state = getattr(component_state, part)
    return component_state


def _get_work_class() -> Callable:
    """Import the work class specified in the environment."""
    work_name = os.environ["LIGHTNING_WORK"]
    work_module_file = os.environ["LIGHTNING_WORK_MODULE_FILE"]
    module = pydoc.importfile(work_module_file)
    return getattr(module, work_name)


def _build_model(work: ServeStreamlit) -> None:
    import streamlit as st

    # Build the model (once per session, equivalent to gradio when enable_queue is Flase)
    if "_model" not in st.session_state:
        with st.spinner("Building model..."):
            st.session_state["_model"] = work.build_model()

    work._model = st.session_state["_model"]


def _main() -> None:
    # Get the AppState
    app_state = AppState(plugin=StreamLitStatePlugin())
    work_state = _reduce_to_component_scope(app_state, os.environ["LIGHTNING_COMPONENT_NAME"])

    # Create the patched work
    work_class = _get_work_class()
    patched_work = _PatchedWork(work_state, work_class)

    # Build and attach the model
    _build_model(patched_work)

    # Render
    patched_work.render()


if __name__ == "__main__":
    _main()



r"""To test a lightning component:

1. Init the component.
2. call .run()
"""
from placeholdername.component import TemplateComponent


def test_placeholder_component():
    messenger = TemplateComponent()
    messenger.run()
    assert messenger.value == 1



# Copyright The PyTorch Lightning team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from typing import List, Optional, Tuple, Union

from lightning.fabric import Fabric
from lightning.fabric.connector import _PLUGIN_INPUT as _LITE_PLUGIN_INPUT
from lightning.fabric.connector import _PRECISION_INPUT
from lightning.fabric.plugins import CheckpointIO, ClusterEnvironment
from lightning.fabric.plugins import DeepSpeedPrecision as LiteDeepSpeedPrecision
from lightning.fabric.plugins import DoublePrecision as LiteDoublePrecision
from lightning.fabric.plugins import MixedPrecision as LiteMixedPrecision
from lightning.fabric.plugins import Precision as LitePrecision
from lightning.fabric.plugins import TPUBf16Precision as LiteTPUBf16Precision
from lightning.fabric.plugins import TPUPrecision as LiteTPUPrecision
from lightning.fabric.strategies import DataParallelStrategy as LiteDataParallelStrategy
from lightning.fabric.strategies import DDPStrategy as LiteDDPStrategy
from lightning.fabric.strategies import DeepSpeedStrategy as LiteDeepSpeedStrategy
from lightning.fabric.strategies import SingleDeviceStrategy as LiteSingleDeviceStrategy
from lightning.fabric.strategies import SingleTPUStrategy as LiteSingleTPUStrategy
from lightning.fabric.strategies import Strategy as LiteStrategy
from lightning.fabric.strategies import XLAStrategy
from lightning.pytorch.accelerators import Accelerator as PLAccelerator
from lightning.pytorch.plugins import DeepSpeedPrecisionPlugin as PLDeepSpeedPrecisionPlugin
from lightning.pytorch.plugins import DoublePrecisionPlugin as PLDoublePrecisionPlugin
from lightning.pytorch.plugins import MixedPrecisionPlugin as PLMixedPrecisionPlugin
from lightning.pytorch.plugins import PrecisionPlugin as PLPrecisionPlugin
from lightning.pytorch.plugins import TPUBf16PrecisionPlugin as PLTPUBf16PrecisionPlugin
from lightning.pytorch.plugins import TPUPrecisionPlugin as PLTPUPrecisionPlugin
from lightning.pytorch.strategies import DataParallelStrategy as PLDataParallelStrategy
from lightning.pytorch.strategies import DDPShardedStrategy as PLDDPShardedStrategy
from lightning.pytorch.strategies import DDPSpawnShardedStrategy as PLDDPSpawnShardedStrategy
from lightning.pytorch.strategies import DDPSpawnStrategy as PLDDPSpawnStrategy
from lightning.pytorch.strategies import DDPStrategy as PLDDPStrategy
from lightning.pytorch.strategies import DeepSpeedStrategy as PLDeepSpeedStrategy
from lightning.pytorch.strategies import SingleDeviceStrategy as PLSingleDeviceStrategy
from lightning.pytorch.strategies import SingleTPUStrategy as PLSingleTPUStrategy
from lightning.pytorch.strategies import Strategy as PLStrategy
from lightning.pytorch.strategies import TPUSpawnStrategy as PLTPUSpawnStrategy
from lightning.pytorch.utilities.rank_zero import rank_zero_deprecation, rank_zero_warn

_PL_PLUGIN = Union[PLPrecisionPlugin, ClusterEnvironment, CheckpointIO]
_PL_PLUGIN_INPUT = Union[_PL_PLUGIN, str]


class LightningLite(Fabric, ABC):
    """Lite accelerates your PyTorch training or inference code with minimal changes required.

    .. deprecated:: v1.9.0
        The `lightning.pytorch.lite.LightningLite` class was deprecated in v1.9.0 and will be renamed to
        `lightning.fabric.Fabric` in v2.0.0. It is no longer part of the pure `lightning.pytorch` package, and now
        lives in the main `lightning` package.

    - Automatic placement of models and data onto the device.
    - Automatic support for mixed and double precision (smaller memory footprint).
    - Seamless switching between hardware (CPU, GPU, TPU) and distributed training strategies
      (data-parallel training, sharded training, etc.).
    - Automated spawning of processes, no launch utilities required.
    - Multi-node support.

    Args:
        accelerator: The hardware to run on. Possible choices are:
            ``"cpu"``, ``"cuda"``, ``"mps"``, ``"gpu"``, ``"tpu"``, ``"auto"``.
        strategy: Strategy for how to run across multiple devices. Possible choices are:
            ``"dp"``, ``"ddp"``, ``"ddp_spawn"``, ``"deepspeed"``, ``"fsdp"``.
        devices: Number of devices to train on (``int``), which GPUs to train on (``list`` or ``str``), or ``"auto"``.
            The value applies per node.
        num_nodes: Number of GPU nodes for distributed training.
        precision: Double precision (``64``), full precision (``32``), half precision (``16``),
            or bfloat16 precision (``"bf16"``).
        plugins: One or several custom plugins
        gpus: Provides the same function as the ``devices`` argument but implies ``accelerator="gpu"``.

            .. deprecated:: v1.8.0
                ``gpus`` has been deprecated in v1.8.0 and will be removed in v2.0.0.
                Please use ``accelerator='gpu'`` and ``devices=x`` instead.

        tpu_cores: Provides the same function as the ``devices`` argument but implies ``accelerator="tpu"``.

            .. deprecated:: v1.8.0
                ``tpu_cores`` has been deprecated in v1.8.0 and will be removed in v2.0.0.
                Please use ``accelerator='tpu'`` and ``devices=x`` instead.
    """

    def __init__(
        self,
        accelerator: Optional[Union[str, PLAccelerator]] = None,
        strategy: Optional[Union[str, PLStrategy]] = None,
        devices: Optional[Union[List[int], str, int]] = None,
        num_nodes: int = 1,
        precision: _PRECISION_INPUT = 32,
        plugins: Optional[Union[_PL_PLUGIN_INPUT, List[_PL_PLUGIN_INPUT]]] = None,
        gpus: Optional[Union[List[int], str, int]] = None,
        tpu_cores: Optional[Union[List[int], str, int]] = None,
    ) -> None:

        rank_zero_deprecation(
            "The `lightning.pytorch.lite.LightningLite` class was deprecated in v1.9.0 and will be renamed to"
            " `lightning.fabric.Fabric` in v2.0.0."
        )

        if gpus is not None or tpu_cores is not None:
            devices, accelerator = _convert_deprecated_device_flags(
                accelerator=accelerator,
                devices=devices,
                gpus=gpus,
                tpu_cores=tpu_cores,
            )

        lite_plugins: Optional[Union[_LITE_PLUGIN_INPUT, List[_LITE_PLUGIN_INPUT]]]
        if isinstance(plugins, PLPrecisionPlugin):
            lite_plugins = _to_lite_precision(plugins)
        elif isinstance(plugins, list):
            lite_plugins = [
                _to_lite_precision(plugin) if isinstance(plugin, PLPrecisionPlugin) else plugin for plugin in plugins
            ]
        else:
            lite_plugins = plugins

        if type(strategy) in (PLDDPShardedStrategy, PLDDPSpawnShardedStrategy) or strategy in (
            "ddp_sharded",
            "ddp_sharded_spawn",
        ):
            spawn_message = ""
            if type(strategy) is PLDDPSpawnShardedStrategy or strategy == "ddp_sharded_spawn":
                spawn_message = ", start_method='spawn'"
            raise RuntimeError(
                "LightningLite's sharded implementation using FairScale has been removed in favor of PyTorch's FSDP."
                " You can try"
                f" `Fabric(strategy=FSDPStrategy(sharding_strategy=ShardingStrategy.SHARD_GRAD_OP{spawn_message}))`"
                " which implements optimizer-only sharding à la ZeRO-2. Or full sharding with `Fabric(strategy='fsdp')`"
            )

        super().__init__(
            accelerator=accelerator,
            strategy=(_to_lite_strategy(strategy) if isinstance(strategy, PLStrategy) else strategy),
            devices=devices,
            num_nodes=num_nodes,
            precision=precision,
            plugins=lite_plugins,
        )


def _convert_deprecated_device_flags(
    accelerator: Optional[Union[str, PLAccelerator]],
    devices: Optional[Union[List[int], str, int]],
    gpus: Optional[Union[List[int], str, int]],
    tpu_cores: Optional[Union[List[int], str, int]],
) -> Tuple[Optional[Union[List[int], str, int]], Optional[Union[str, PLAccelerator]]]:
    """Emit deprecation warnings for gpus and tpu_cores and translate them into the new accelerator and devices.

    Similar implementation as in ``lightning.pytorch.trainer.connectors.accelerator_connector``.
    """
    if gpus is not None:
        rank_zero_deprecation(
            f"Setting `Lite(gpus={gpus!r})` is deprecated in v1.8.0 and will be removed"
            f" in v2.0.0. Please use `Lite(accelerator='gpu', devices={gpus!r})` instead."
        )
    if tpu_cores is not None:
        rank_zero_deprecation(
            f"Setting `Lite(tpu_cores={tpu_cores!r})` is deprecated in v1.8.0 and will be removed"
            f" in v2.0.0. Please use `Lite(accelerator='tpu', devices={tpu_cores!r})` instead."
        )
    deprecated_devices_specific_flag = gpus or tpu_cores
    if deprecated_devices_specific_flag and deprecated_devices_specific_flag not in ([], 0, "0"):
        if devices:
            rank_zero_warn(
                f"The option `devices={devices}` will be ignored and the device specific number"
                f"{deprecated_devices_specific_flag} will be used instead."
            )

        if gpus is not None and tpu_cores is not None:
            rank_zero_warn(
                f"Both `Lite(gpus={gpus!r}, tpu_cores={tpu_cores!r})` were specified. Please choose only one of"
                " the two."
            )

        if accelerator is None:
            if tpu_cores:
                accelerator = "tpu"
            if gpus:
                accelerator = "cuda"

    return deprecated_devices_specific_flag, accelerator


def _to_lite_strategy(strategy: PLStrategy) -> LiteStrategy:
    """Re-instantiates a PL-Strategy as the corresponding Lite-Strategy."""
    strategy_cls = type(strategy)
    if strategy_cls is PLDDPStrategy:
        return LiteDDPStrategy(
            accelerator=strategy.accelerator,
            parallel_devices=strategy.parallel_devices,
            cluster_environment=strategy.cluster_environment,
            checkpoint_io=strategy.checkpoint_io,
            precision=_to_lite_precision(strategy.precision_plugin),
            process_group_backend=strategy.process_group_backend,
            timeout=strategy._timeout,
            **strategy._ddp_kwargs,
        )

    if strategy_cls is PLDDPSpawnStrategy:
        return LiteDDPStrategy(
            accelerator=strategy.accelerator,
            parallel_devices=strategy.parallel_devices,
            cluster_environment=strategy.cluster_environment,
            checkpoint_io=strategy.checkpoint_io,
            precision=_to_lite_precision(strategy.precision_plugin),
            process_group_backend=strategy.process_group_backend,
            timeout=strategy._timeout,
            start_method=strategy._start_method,
            **strategy._ddp_kwargs,
        )

    if strategy_cls is PLTPUSpawnStrategy:
        return XLAStrategy(
            accelerator=strategy.accelerator,
            parallel_devices=strategy.parallel_devices,
            checkpoint_io=strategy.checkpoint_io,
            precision=_to_lite_precision(strategy.precision_plugin),
        )

    if strategy_cls is PLDeepSpeedStrategy:
        return LiteDeepSpeedStrategy(
            accelerator=strategy.accelerator,
            parallel_devices=strategy.parallel_devices,
            cluster_environment=strategy.cluster_environment,
            precision=_to_lite_precision(strategy.precision_plugin),
            process_group_backend=strategy.process_group_backend,
            config=strategy.config,
            remote_device=strategy.remote_device,
            load_full_weights=strategy.load_full_weights,
            loss_scale=strategy.loss_scale,
            initial_scale_power=strategy.initial_scale_power,
            loss_scale_window=strategy.loss_scale_window,
            hysteresis=strategy.hysteresis,
            min_loss_scale=strategy.min_loss_scale,
        )

    if strategy_cls is PLDataParallelStrategy:
        return LiteDataParallelStrategy(
            accelerator=strategy.accelerator,
            parallel_devices=strategy.parallel_devices,
            checkpoint_io=strategy.checkpoint_io,
            precision=_to_lite_precision(strategy.precision_plugin),
        )

    if strategy_cls is PLSingleDeviceStrategy:
        return LiteSingleDeviceStrategy(
            device=strategy.root_device,
            accelerator=strategy.accelerator,
            checkpoint_io=strategy.checkpoint_io,
            precision=_to_lite_precision(strategy.precision_plugin),
        )

    if strategy_cls is PLSingleTPUStrategy:
        return LiteSingleTPUStrategy(
            device=strategy.root_device.index,
            accelerator=strategy.accelerator,
            checkpoint_io=strategy.checkpoint_io,
            precision=_to_lite_precision(strategy.precision_plugin),
        )
    raise NotImplementedError(f"Unsupported strategy: `{strategy_cls.__name__}`")


def _to_lite_precision(plugin: Optional[PLPrecisionPlugin]) -> LitePrecision:
    """Re-instantiates a PL-PrecisionPlugin as the corresponding Lite-Precision plugin."""

    if type(plugin) is PLPrecisionPlugin:
        return LitePrecision()

    if type(plugin) is PLMixedPrecisionPlugin:
        return LiteMixedPrecision(
            precision=plugin.precision, device=plugin.device, scaler=plugin.scaler  # type: ignore[arg-type]
        )

    if type(plugin) is PLDoublePrecisionPlugin:
        return LiteDoublePrecision()

    if type(plugin) is PLDeepSpeedPrecisionPlugin:
        return LiteDeepSpeedPrecision(
            precision=plugin.precision,  # type: ignore[arg-type]
        )

    if type(plugin) is PLTPUPrecisionPlugin:
        return LiteTPUPrecision()

    if type(plugin) is PLTPUBf16PrecisionPlugin:
        return LiteTPUBf16Precision()

    # No backward compatibility for custom plugins / subclasses, as we can't re-instantiate these plugins
    raise TypeError(
        "You passed an unsupported plugin as input to Lite(plugins=...) or to a strategy. If you built a custom plugin,"
        " please change it to subclass the `lightning_lite.plugins.precision.Precision` class. Otherwise, please open"
        " an issue on the Lightning GitHub repository with your use case."
    )



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import inspect
import time
from copy import deepcopy
from dataclasses import dataclass
from functools import wraps
from multiprocessing import Queue
from typing import Any, Callable, Dict, List, Optional
from uuid import uuid4

from fastapi import FastAPI, HTTPException, Request, status
from lightning_utilities.core.apply_func import apply_to_collection

from lightning.app.api.request_types import _APIRequest, _CommandRequest, _RequestResponse
from lightning.app.utilities.app_helpers import Logger

logger = Logger(__name__)


def _signature_proxy_function():
    pass


@dataclass
class _FastApiMockRequest:
    """This class is meant to mock FastAPI Request class that isn't pickle-able.

    If a user relies on FastAPI Request annotation, the Lightning framework
    patches the annotation before pickling and replace them right after.

    Finally, the FastAPI request is converted back to the _FastApiMockRequest
    before being delivered to the users.

    Example:

        import lightning as L
        from fastapi import Request
        from lightning.app.api import Post

        class Flow(L.LightningFlow):

            def request(self, request: Request) -> OutputRequestModel:
                ...

            def configure_api(self):
                return [Post("/api/v1/request", self.request)]
    """

    _body: Optional[str] = None
    _json: Optional[str] = None
    _method: Optional[str] = None
    _headers: Optional[Dict] = None

    @property
    def receive(self):
        raise NotImplementedError

    @property
    def method(self):
        raise self._method

    @property
    def headers(self):
        return self._headers

    def body(self):
        return self._body

    def json(self):
        return self._json

    def stream(self):
        raise NotImplementedError

    def form(self):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError

    def is_disconnected(self):
        raise NotImplementedError


async def _mock_fastapi_request(request: Request):
    # TODO: Add more requests parameters.
    return _FastApiMockRequest(
        _body=await request.body(),
        _json=await request.json(),
        _headers=request.headers,
        _method=request.method,
    )


class _HttpMethod:
    def __init__(self, route: str, method: Callable, method_name: Optional[str] = None, timeout: int = 30, **kwargs):
        """This class is used to inject user defined methods within the App Rest API.

        Arguments:
            route: The path used to route the requests
            method: The associated flow method
            timeout: The time in seconds taken before raising a timeout exception.
        """
        self.route = route
        self.attached_to_flow = hasattr(method, "__self__")
        self.method_name = method_name or method.__name__
        self.method_annotations = method.__annotations__
        # TODO: Validate the signature contains only pydantic models.
        self.method_signature = inspect.signature(method)

        if not self.attached_to_flow:
            self.component_name = method.__name__
            self.method = method
        else:
            self.component_name = method.__self__.name

        self.timeout = timeout
        self.kwargs = kwargs

        # Enable the users to rely on FastAPI annotation typing with Request.
        # Note: Only a part of the Request functionatilities are supported.
        self._patch_fast_api_request()

    def add_route(self, app: FastAPI, request_queue: Queue, responses_store: Dict[str, Any]) -> None:
        # 1: Get the route associated with the http method.
        route = getattr(app, self.__class__.__name__.lower())

        self._unpatch_fast_api_request()

        # 2: Create a proxy function with the signature of the wrapped method.
        fn = deepcopy(_signature_proxy_function)
        fn.__annotations__ = self.method_annotations
        fn.__name__ = self.method_name
        setattr(fn, "__signature__", self.method_signature)

        # Note: Handle requests differently if attached to a flow.
        if not self.attached_to_flow:
            # 3: Define the request handler.
            @wraps(_signature_proxy_function)
            async def _handle_request(*args, **kwargs):
                if inspect.iscoroutinefunction(self.method):
                    return await self.method(*args, **kwargs)
                return self.method(*args, **kwargs)

        else:
            request_cls = _CommandRequest if self.route.startswith("/command/") else _APIRequest

            # 3: Define the request handler.
            @wraps(_signature_proxy_function)
            async def _handle_request(*args, **kwargs):
                async def fn(*args, **kwargs):
                    args, kwargs = apply_to_collection((args, kwargs), Request, _mock_fastapi_request)
                    for k, v in kwargs.items():
                        if hasattr(v, "__await__"):
                            kwargs[k] = await v

                    request_id = str(uuid4()).split("-")[0]
                    logger.debug(f"Processing request {request_id} for route: {self.route}")
                    request_queue.put(
                        request_cls(
                            name=self.component_name,
                            method_name=self.method_name,
                            args=args,
                            kwargs=kwargs,
                            id=request_id,
                        )
                    )

                    t0 = time.time()
                    while request_id not in responses_store:
                        await asyncio.sleep(0.01)
                        if (time.time() - t0) > self.timeout:
                            raise HTTPException(
                                status.HTTP_500_INTERNAL_SERVER_ERROR,
                                detail="The response was never received.",
                            )

                    logger.debug(f"Processed request {request_id} for route: {self.route}")

                    return responses_store.pop(request_id)

                response: _RequestResponse = await asyncio.create_task(fn(*args, **kwargs))

                if response.status_code != 200:
                    raise HTTPException(response.status_code, detail=response.content)

                return response.content

        # 4: Register the user provided route to the Rest API.
        route(self.route, **self.kwargs)(_handle_request)

    def _patch_fast_api_request(self):
        """This function replaces signature annotation for Request with its mock."""
        for k, v in self.method_annotations.items():
            if v == Request:
                self.method_annotations[k] = _FastApiMockRequest

        for v in self.method_signature.parameters.values():
            if v._annotation == Request:
                v._annotation = _FastApiMockRequest

    def _unpatch_fast_api_request(self):
        """This function replaces back signature annotation to fastapi Request."""
        for k, v in self.method_annotations.items():
            if v == _FastApiMockRequest:
                self.method_annotations[k] = Request

        for v in self.method_signature.parameters.values():
            if v._annotation == _FastApiMockRequest:
                v._annotation = Request


class Post(_HttpMethod):
    pass


class Get(_HttpMethod):
    pass


class Put(_HttpMethod):
    pass


class Delete(_HttpMethod):
    pass


def _add_tags_to_api(apis: List[_HttpMethod], tags: List[str]) -> None:
    for api in apis:
        if not api.kwargs.get("tag"):
            api.kwargs["tags"] = tags


def _validate_api(apis: List[_HttpMethod]) -> None:
    for api in apis:
        if not isinstance(api, _HttpMethod):
            raise Exception(f"The provided api should be either [{Delete}, {Get}, {Post}, {Put}]")
        if api.route.startswith("/command"):
            raise Exception("The route `/command` is reserved for commands. Please, use something else.")
